# Copyright (C) 2020-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

# Co-authored-by: Brandon Edwards <brandon.edwards@intel.com>
# Co-authored-by: Anindya S. Paul <anindya.s.paul@intel.com>
# Co-authored-by: Mansi Sharma <mansi.sharma@intel.com>

from clip_optimizer import ClipOptimizer
from copy import deepcopy
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torch
import torchvision
from torch.utils.data import TensorDataset
import numpy as np
from openfl.experimental.workflow.interface import FLSpec, Aggregator, Collaborator
from openfl.experimental.workflow.runtime import LocalRuntime
from openfl.experimental.workflow.placement import aggregator, collaborator

from opacus import PrivacyEngine
import argparse
import yaml

import warnings

warnings.filterwarnings("ignore")


batch_size_train = 64
batch_size_test = 1000
learning_rate = 0.01
momentum = 0.5
log_interval = 10

random_seed = 5495300300540669060

g_device = torch.Generator(device="cuda")
# Uncomment the line below to use g_cpu if not using cuda
# g_device = torch.Generator() # noqa: E800
# NOTE: remove below to stop repeatable runs
g_device.manual_seed(random_seed)
print(f"\n\nWe are using seed: {random_seed}")

mnist_train = torchvision.datasets.MNIST(
    "files/",
    train=True,
    download=True,
    transform=torchvision.transforms.Compose(
        [
            torchvision.transforms.ToTensor(),
            torchvision.transforms.Normalize((0.1307,), (0.3081,)),
        ]
    ),
)

mnist_test = torchvision.datasets.MNIST(
    "files/",
    train=False,
    download=True,
    transform=torchvision.transforms.Compose(
        [
            torchvision.transforms.ToTensor(),
            torchvision.transforms.Normalize((0.1307,), (0.3081,)),
        ]
    ),
)


class GlobalModelTools(object):
    """
    Class to facilitate the updating of a model whose 'per-sample' delta updates
    are provided from some out of band process and passed as torch model
    state dicts to the class. A global model object is constructed and its
    per example gradients are filled in with the local model updates. Then the
    optimizer can be stepped to affect the aggregate update (with DP if desired)
    """

    def __init__(
        self, example_model_state, global_model, collaborator_names, dp_params=None
    ):
        self.example_state = example_model_state
        self.global_model = global_model
        self.collaborator_names = collaborator_names
        self.global_optimizer = torch.optim.SGD(
            params=self.global_model.parameters(), lr=1.0
        )  # This choice of optimizer is required for correct model aggregation
        self.dp_params = dp_params
        self.privacy_engine = PrivacyEngine()
        if dp_params is None:
            # this is the expected fraction of collaborators to select each round,
            # also the indepencent probability that each
            # collaborator gets selected for a round
            sample_rate = 1.0
        else:
            sample_rate = dp_params["sample_rate"]
        self.global_data_loader = torch.utils.data.DataLoader(
            TensorDataset(
                torch.Tensor(list(range(len(self.collaborator_names)))).to(torch.int)
            ),
            batch_size=int(sample_rate * float(len(collaborator_names))),
            shuffle=True,
        )
        if dp_params is not None:
            (
                self.global_model,
                self.global_optimizer,
                self.global_data_loader,
            ) = self.privacy_engine.make_private(
                module=self.global_model,
                optimizer=self.global_optimizer,
                data_loader=self.global_data_loader,
                noise_multiplier=dp_params["noise_multiplier"],
                max_grad_norm=dp_params["clip_norm"],
            )

    def populate_model_params_and_gradients(
        self, state_for_params, states_for_gradients
    ):
        if self.dp_params is not None:
            # populate param.grad_sample with individual updates from state_dicts,
            # and populate pram.grad with a random
            # properly shaped update in order to help opacus infer that shape information
            per_layer_grad_samples = {
                key: [state[key] for state in states_for_gradients]
                for key in states_for_gradients[0].keys()
            }
            for params, name in zip(
                self.global_model.parameters(), self.example_state.keys()
            ):
                params.data = state_for_params[name]
                params.grad_sample = torch.stack(per_layer_grad_samples[name], dim=0)
                # only the shape is important below, values are not important and so we
                # use only the first state
                params.grad = states_for_gradients[0][name]


class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 10, kernel_size=5)
        self.conv2 = nn.Conv2d(10, 20, kernel_size=5)
        self.conv2_drop = nn.Dropout2d()
        self.fc1 = nn.Linear(320, 50)
        self.fc2 = nn.Linear(50, 10)

    def forward(self, x):
        x = F.relu(F.max_pool2d(self.conv1(x), 2))
        x = F.relu(F.max_pool2d(self.conv2_drop(self.conv2(x)), 2))
        x = x.view(-1, 320)
        x = F.relu(self.fc1(x))
        x = F.dropout(x, training=self.training)
        x = self.fc2(x)
        return F.log_softmax(x)


def default_optimizer(model):
    # indicate the default optimizer for training the target model
    return optim.SGD(model.parameters(), lr=learning_rate, momentum=momentum)


def FedAvg(models, global_model_tools, previous_global_state, dp_params):  # NOQA: N802
    """
    This tutorial utilizes non-weighted averaging of collaborator model updates
    regardless of whether DP config is used.
    Weighted FedAvg is currently not supported.
    """
    new_model = models[0]
    # Note: the states below are non-delta states
    non_delta_states = [model.state_dict() for model in models]
    neg_delta_states = []
    for non_delta_state in non_delta_states:
        # These will be used for the gradient, so need to be opposite the model delta
        neg_delta_states.append(
            {
                key: -non_delta_state[key] + previous_global_state[key]
                for key in non_delta_state
            }
        )
    # validate that in fact the local models clipped their updates
    print()
    for idx, state in enumerate(neg_delta_states):
        per_layer_norms = []
        for key, tensor in state.items():
            per_layer_norms.append(torch.norm(tensor, dim=()))
        delta_norm = torch.norm(torch.Tensor(per_layer_norms), dim=())
        print(f"delta_norm for idx {idx} is: {delta_norm}")
        if delta_norm > dp_params["clip_norm"]:
            raise ValueError(
                f"The model with index {idx} had update whose "
                + "L2-norm was greater than clip norm. Correct the local periodic clipping."
            )
    print()
    # Clearing state from optimizer from last round so as to not leak information
    global_model_tools.global_optimizer.zero_grad()
    global_model_tools.populate_model_params_and_gradients(
        state_for_params=previous_global_state, states_for_gradients=neg_delta_states
    )
    global_model_tools.global_optimizer.step()

    # removing the '_module.' from the beginning of all keys coming from
    # global_model_tools.global_model state dict
    new_model.load_state_dict(
        {
            key[8:]: value
            for key, value in global_model_tools.global_model.state_dict().items()
        }
    )
    return new_model


def inference(network, test_loader, device):
    network.eval()
    test_loss = 0
    correct = 0
    with torch.no_grad():
        for data, target in test_loader:
            data = data.to(device)
            target = target.to(device)
            output = network(data)
            test_loss += F.nll_loss(output, target, size_average=False).item()
            pred = output.data.max(1, keepdim=True)[1]
            correct += pred.eq(target.data.view_as(pred)).sum()
    test_loss /= len(test_loader.dataset)
    print(
        "\nTest set: Avg. loss: {test_loss:.4f},"
        f" Accuracy: {correct}/{len(test_loader.dataset)}"
        f" ({(100.0 * correct / len(test_loader.dataset)):.0f})\n"
    )
    accuracy = float(correct / len(test_loader.dataset))
    return accuracy


def optimizer_to_device(optimizer, device):
    if optimizer.state_dict != {}:
        if isinstance(optimizer, optim.SGD):
            for param in optimizer.param_groups[0]["params"]:
                param.data = param.data.to(device)
                if param.grad is not None:
                    param.grad = param.grad.to(device)
        elif isinstance(optimizer, optim.Adam):
            for state in optimizer.state.values():
                for k, v in state.items():
                    if isinstance(v, torch.Tensor):
                        state[k] = v.to(device)
    else:
        raise (
            ValueError("Current optimizer state does not have dict keys: please verify")
        )


def clip_testing_on_optimizer_parameters(
    optimizer_before_step_params,
    optimizer_after_step_params,
    collaborator_name,
    round_num,
    device,
):
    len_equal_tensor = 0
    for param_idx in range(len(optimizer_after_step_params)):
        for tensor_1, tensor_2 in zip(
            optimizer_before_step_params[param_idx],
            optimizer_after_step_params[param_idx],
        ):
            if torch.equal(tensor_1.to(device), tensor_2.to(device)) is True:
                len_equal_tensor += 1
    if len_equal_tensor == len(optimizer_after_step_params):
        raise (
            ValueError(
                "No clipping effect: Optimizer param data is the same "
                + "between before and after optimizer step for collaborator: "
                + f"{collaborator_name} at round {round_num}"
            )
        )


def validate_dp_params(dp_params):
    # The differential privacy block should have the exact keys provided below
    required_dp_keys = [
        "clip_norm",
        "noise_multiplier",
        "delta",
        "sample_rate",
        "clip_frequency",
    ]

    keys = dp_params.keys()
    excess = list(set(keys).difference(set(required_dp_keys)))
    deficit = list(set(required_dp_keys).difference(keys))

    if excess != []:
        print(
            f"\nCAUTION: The keys: {excess} where provided in the 'differential_privacy'"
            + " block of the flow config and are not being used.\n"
        )
    if deficit != []:
        raise ValueError(
            f"The 'differential_privacy' block is missing the required keys: {deficit}"
        )


def parse_config(config_path):
    with open(config_path, "rb") as _file:
        config = yaml.safe_load(_file)
    return config


class FederatedFlow(FLSpec):
    def __init__(
        self,
        config_path,
        model,
        collaborator_names,
        device,
        total_rounds=10,
        top_model_accuracy=0,
        clip_test=False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.model = model
        self.global_model = Net()
        self.collaborator_names = collaborator_names
        self.total_rounds = total_rounds
        self.round = None
        self.top_model_accuracy = top_model_accuracy
        self.device = device
        self.clip_test = clip_test
        # we will set this attribute at the beginning of each round (tracks using
        # indices against the collaborator list)
        self.round_collaborator_idxs = None

        config = parse_config(config_path)

        if "differential_privacy" not in config:
            self.dp_params = None
        else:
            self.dp_params = config["differential_privacy"]
            print(f"Here are dp_params: {self.dp_params}")
            validate_dp_params(self.dp_params)
        self.global_model_tools = GlobalModelTools(
            global_model=self.global_model,
            example_model_state=self.model.state_dict(),
            collaborator_names=self.collaborator_names,
            dp_params=self.dp_params,
        )

    @aggregator
    def start(self):
        print("Performing initialization for model")
        self.collaborators = self.runtime.collaborators
        self.private = 10

        # determine starting round collaborators by grabbing first batch of loader
        round_collaborator_idxs = [
            batch[0] for batch in self.global_model_tools.global_data_loader
        ][0]
        self.round_collaborators = [
            self.collaborator_names[idx] for idx in round_collaborator_idxs
        ]

        if self.round is None:
            self.round = 0
            if self.dp_params is None:
                self.optimizers = {
                    collaborator_name: default_optimizer(self.model)
                    for collaborator_name in self.collaborator_names
                }
            else:
                self.optimizers = {
                    collaborator_name: ClipOptimizer(
                        base_optimizer=default_optimizer(self.model),
                        device=self.device,
                        clip_norm=self.dp_params["clip_norm"],
                        clip_freq=self.dp_params["clip_frequency"],
                    )
                    for collaborator_name in self.collaborator_names
                }

        print("\n\n" + 20 * "#")
        print("Round {self.round}...")
        print("Training with collaborators: ", self.round_collaborators)
        print(20 * "#" + "\n\n")

        if (len(round_collaborator_idxs) != 0) and (
            round_collaborator_idxs[0].nelement() != 0
        ):
            self.next(
                self.aggregated_model_validation,
                foreach="round_collaborators",
                exclude=["private"],
            )
        else:
            if self.round == self.total_rounds - 1:
                print(
                    "Completed all rounds with no collaborator selected for training at any round"
                )
                self.next(self.end)

            self.round += 1
            self.next(self.start)

    @collaborator
    def aggregated_model_validation(self):
        print(f"Performing aggregated model validation for collaborator {self.input}")
        self.model = self.model.to(self.device)
        self.global_model = self.global_model.to(self.device)

        # verifying that model went to the correct GPU device
        assert next(self.model.parameters()).device == self.device
        assert next(self.global_model.parameters()).device == self.device

        self.agg_validation_score = inference(self.model, self.test_loader, self.device)
        print(f"{self.input} value of {self.agg_validation_score}")
        self.collaborator_name = self.input
        self.next(self.train)

    @collaborator
    def train(self):
        print(f"Performing model training for collaborator {self.input}")

        self.optimizer = ClipOptimizer(
            base_optimizer=default_optimizer(self.model),
            device=self.device,
            clip_norm=self.dp_params["clip_norm"],
            clip_freq=self.dp_params["clip_frequency"],
        )

        # copy state dict from optimizer object from previous round to the newly
        # instantiated optimizer for the same collaborator
        if self.round > 0:
            optimizer_to_device(optimizer=self.optimizer, device=self.device)
            self.optimizer.load_state_dict(
                deepcopy(self.optimizers[self.input].state_dict())
            )

        self.model.train()
        train_losses = []
        for batch_idx, (data, target) in enumerate(self.train_loader):
            data = data.to(self.device)
            target = target.to(self.device)
            self.optimizer.zero_grad()
            output = self.model(data)
            loss = F.nll_loss(output, target).to(self.device)
            loss.backward()

            if self.clip_test:
                optimizer_before_step_params = [
                    param.data for param in self.optimizer.param_groups()[0]["params"]
                ]

            self.optimizer.step(
                global_model_state=self.global_model.state_dict(),
                last_iter=(batch_idx == (len(self.train_loader) - 1)),
            )

            if batch_idx % self.dp_params["clip_frequency"] == 0 or (
                batch_idx == (len(self.train_loader) - 1)
            ):
                if self.clip_test:
                    optimizer_after_step_params = [
                        param.data
                        for param in self.optimizer.param_groups()[0]["params"]
                    ]
                    clip_testing_on_optimizer_parameters(
                        optimizer_before_step_params,
                        optimizer_after_step_params,
                        self.collaborator_name,
                        self.round,
                        self.device,
                    )

            train_losses.append(loss.item())

        self.loss = np.mean(train_losses)
        self.training_completed = True

        tmp_opt = deepcopy(self.optimizers[self.input])
        tmp_opt.load_state_dict(self.optimizer.state_dict())
        self.optimizer = tmp_opt
        torch.cuda.empty_cache()
        self.next(self.local_model_validation)

    @collaborator
    def local_model_validation(self):
        print(f"Performing local model validation for collaborator {self.input}")
        self.local_validation_score = inference(
            self.model, self.test_loader, self.device
        )
        print(f"{self.input} value of {self.local_validation_score}")
        self.next(self.join, exclude=["training_completed"])

    @aggregator
    def join(self, inputs):
        self.average_loss = sum(input.loss for input in inputs) / len(inputs)
        self.aggregated_model_accuracy = sum(
            input.agg_validation_score for input in inputs
        ) / len(inputs)
        self.local_model_accuracy = sum(
            input.local_validation_score for input in inputs
        ) / len(inputs)
        print(
            f"Average aggregated model validation values = {self.aggregated_model_accuracy}"
        )
        print(f"Average training loss = {self.average_loss}")
        print(f"Average local model validation values = {self.local_model_accuracy}")
        self.model.load_state_dict(
            FedAvg(
                [input.model.cpu() for input in inputs],
                global_model_tools=self.global_model_tools,
                previous_global_state=self.global_model.cpu().state_dict(),
                dp_params=self.dp_params,
            ).state_dict()
        )

        if self.dp_params is not None:
            print(15 * "#")
            epsilon = self.global_model_tools.privacy_engine.get_epsilon(
                delta=self.dp_params["delta"]
            )
            print(
                "\nCurrent epsilon is: "
                + f"{epsilon} for delta: {self.dp_params['delta']}"
            )
            print(15 * "#")

        self.global_model.load_state_dict(deepcopy(self.model.state_dict()))
        self.optimizers.update(
            {input.collaborator_name: input.optimizer for input in inputs}
        )

        del inputs
        self.next(self.check_round_completion)

    @aggregator
    def check_round_completion(self):
        if self.round == self.total_rounds - 1:
            self.next(self.end)
        else:
            if self.aggregated_model_accuracy > self.top_model_accuracy:
                print(
                    f"Accuracy improved to {self.aggregated_model_accuracy}"
                    + f"for round {self.round}"
                )
                self.top_model_accuracy = self.aggregated_model_accuracy
            self.round += 1

            # determine starting round collaborators by grabbing first batch of loader
            for batch in self.global_model_tools.global_data_loader:
                round_collaborator_idxs = batch
                break
            self.round_collaborators = [
                self.collaborator_names[idx] for idx in round_collaborator_idxs[0]
            ]

            print("\n\n" + 20 * "#")
            print(f"Round {self.round}...")
            print("Training with collaborators: ", self.round_collaborators)
            print(20 * "#" + "\n\n")

            if (len(round_collaborator_idxs) != 0) and (
                round_collaborator_idxs[0].nelement() != 0
            ):
                self.next(
                    self.aggregated_model_validation,
                    foreach="round_collaborators",
                    exclude=["private"],
                )
            else:
                self.next(self.check_round_completion)

    @aggregator
    def end(self):
        print(20 * "#")
        print("All rounds completed successfully")
        print(20 * "#")
        print("This is the end of the flow")
        print(20 * "#")


if __name__ == "__main__":
    argparser = argparse.ArgumentParser(description=__doc__)
    argparser.add_argument(
        "--config_path", help="Absolute path to the flow configuration file."
    )
    argparser.add_argument(
        "--clip_test",
        default=False,
        help="Indicate enabling of optimizer param testing before and after clip",
    )

    args = argparser.parse_args()

    if torch.cuda.is_available():
        device = torch.device("cuda:0")
    else:
        device = torch.device("cpu")

    # Setup participants
    # Set `num_gpus=0.09` to `num_gpus=0.0` in order to run this tutorial on CPU
    aggregator = Aggregator(num_gpus=0.09)

    # Setup collaborators with private attributes
    collaborator_names = [
        "Portland",
        "Seattle",
        "Bangalore",
        "Chandler",
        "Austin",
        "Hudson",
        "Penang",
        "Haifa",
        "CostaRica",
        "Guadalajara",
    ]

    def callable_to_initialize_collaborator_private_attributes(
        index, n_collaborators, batch_size, train_dataset, test_dataset
    ):
        train = deepcopy(train_dataset)
        test = deepcopy(test_dataset)
        train.data = train_dataset.data[index::n_collaborators]
        train.targets = train_dataset.targets[index::n_collaborators]
        test.data = test_dataset.data[index::n_collaborators]
        test.targets = test_dataset.targets[index::n_collaborators]

        return {
            "train_loader": torch.utils.data.DataLoader(
                train, batch_size=batch_size, shuffle=True
            ),
            "test_loader": torch.utils.data.DataLoader(
                test, batch_size=batch_size, shuffle=True
            ),
        }

    collaborators = []
    for idx, collaborator_name in enumerate(collaborator_names):
        collaborators.append(
            Collaborator(
                name=collaborator_name,
                private_attributes_callable=callable_to_initialize_collaborator_private_attributes,
                # Set `num_gpus=0.09` to `num_gpus=0.0` in order to run this tutorial on CPU
                num_cpus=0.0,
                num_gpus=0.09,  # Assuming GPU(s) is available in the machine
                index=idx,
                n_collaborators=len(collaborator_names),
                batch_size=batch_size_train,
                train_dataset=mnist_train,
                test_dataset=mnist_test,
            )
        )

    local_runtime = LocalRuntime(
        aggregator=aggregator, collaborators=collaborators, backend="ray"
    )
    print(f"Local runtime collaborators = {local_runtime.collaborators}")

    best_model = None
    initial_model = Net()
    top_model_accuracy = 0
    total_rounds = 10

    flflow = FederatedFlow(
        config_path=args.config_path,
        model=initial_model,
        collaborator_names=collaborator_names,
        device=device,
        total_rounds=total_rounds,
        top_model_accuracy=top_model_accuracy,
        clip_test=args.clip_test,
    )
    flflow.runtime = local_runtime
    flflow.run()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: openfl/experimental/workflow/protocols/director.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n5openfl/experimental/workflow/protocols/director.proto\x12%openfl.experimental.workflow.director\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1egoogle/protobuf/duration.proto\"+\n\x15SendConnectionRequest\x12\x12\n\nenvoy_name\x18\x01 \x01(\t\"#\n\x0fRequestAccepted\x12\x10\n\x08\x61\x63\x63\x65pted\x18\x01 \x01(\x08\"2\n\x15WaitExperimentRequest\x12\x19\n\x11\x63ollaborator_name\x18\x01 \x01(\t\"1\n\x16WaitExperimentResponse\x12\x17\n\x0f\x65xperiment_name\x18\x01 \x01(\t\"N\n\x18GetExperimentDataRequest\x12\x17\n\x0f\x65xperiment_name\x18\x01 \x01(\t\x12\x19\n\x11\x63ollaborator_name\x18\x02 \x01(\t\"0\n\x0e\x45xperimentData\x12\x0c\n\x04size\x18\x01 \x01(\r\x12\x10\n\x08\x65xp_data\x18\x02 \x01(\x0c\"G\n\x18UpdateEnvoyStatusRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1d\n\x15is_experiment_running\x18\x02 \x01(\x08\"S\n\x19UpdateEnvoyStatusResponse\x12\x36\n\x13health_check_period\x18\x01 \x01(\x0b\x32\x19.google.protobuf.Duration\"\x8a\x01\n\x0e\x45xperimentInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1a\n\x12\x63ollaborator_names\x18\x02 \x03(\t\x12N\n\x0f\x65xperiment_data\x18\x03 \x01(\x0b\x32\x35.openfl.experimental.workflow.director.ExperimentData\"*\n\x18SetNewExperimentResponse\x12\x0e\n\x06status\x18\x01 \x01(\x08\"\xcf\x01\n\tEnvoyInfo\x12\x12\n\nenvoy_name\x18\x01 \x01(\t\x12\x17\n\x0f\x65xperiment_name\x18\x02 \x01(\t\x12\x11\n\tis_online\x18\x03 \x01(\x08\x12\x1d\n\x15is_experiment_running\x18\x04 \x01(\x08\x12\x30\n\x0clast_updated\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x31\n\x0evalid_duration\x18\x06 \x01(\x0b\x32\x19.google.protobuf.Duration\"\x12\n\x10GetEnvoysRequest\"Z\n\x11GetEnvoysResponse\x12\x45\n\x0b\x65nvoy_infos\x18\x01 \x03(\x0b\x32\x30.openfl.experimental.workflow.director.EnvoyInfo\"\x15\n\x13GetFlowStateRequest\"=\n\x14GetFlowStateResponse\x12\x11\n\tcompleted\x18\x01 \x01(\x08\x12\x12\n\nflspec_obj\x18\x02 \x01(\x0c\"\x14\n\x12SendRuntimeRequest\"*\n\x16RuntimeRequestResponse\x12\x10\n\x08\x61\x63\x63\x65pted\x18\x01 \x01(\x08\"5\n\x1aGetExperimentStdoutRequest\x12\x17\n\x0f\x65xperiment_name\x18\x01 \x01(\t\"]\n\x1bGetExperimentStdoutResponse\x12\x15\n\rstdout_origin\x18\x01 \x01(\t\x12\x11\n\ttask_name\x18\x02 \x01(\t\x12\x14\n\x0cstdout_value\x18\x03 \x01(\t2\xae\n\n\x08\x44irector\x12\x90\x01\n\x16\x45nvoyConnectionRequest\x12<.openfl.experimental.workflow.director.SendConnectionRequest\x1a\x36.openfl.experimental.workflow.director.RequestAccepted\"\x00\x12\x8f\x01\n\x0eWaitExperiment\x12<.openfl.experimental.workflow.director.WaitExperimentRequest\x1a=.openfl.experimental.workflow.director.WaitExperimentResponse\"\x00\x12\x8f\x01\n\x11GetExperimentData\x12?.openfl.experimental.workflow.director.GetExperimentDataRequest\x1a\x35.openfl.experimental.workflow.director.ExperimentData\"\x00\x30\x01\x12\x98\x01\n\x11UpdateEnvoyStatus\x12?.openfl.experimental.workflow.director.UpdateEnvoyStatusRequest\x1a@.openfl.experimental.workflow.director.UpdateEnvoyStatusResponse\"\x00\x12\x8e\x01\n\x10SetNewExperiment\x12\x35.openfl.experimental.workflow.director.ExperimentInfo\x1a?.openfl.experimental.workflow.director.SetNewExperimentResponse\"\x00(\x01\x12\x80\x01\n\tGetEnvoys\x12\x37.openfl.experimental.workflow.director.GetEnvoysRequest\x1a\x38.openfl.experimental.workflow.director.GetEnvoysResponse\"\x00\x12\x89\x01\n\x0cGetFlowState\x12:.openfl.experimental.workflow.director.GetFlowStateRequest\x1a;.openfl.experimental.workflow.director.GetFlowStateResponse\"\x00\x12\x8c\x01\n\x0e\x43onnectRuntime\x12\x39.openfl.experimental.workflow.director.SendRuntimeRequest\x1a=.openfl.experimental.workflow.director.RuntimeRequestResponse\"\x00\x12\xa0\x01\n\x13GetExperimentStdout\x12\x41.openfl.experimental.workflow.director.GetExperimentStdoutRequest\x1a\x42.openfl.experimental.workflow.director.GetExperimentStdoutResponse\"\x00\x30\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'openfl.experimental.workflow.protocols.director_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_SENDCONNECTIONREQUEST']._serialized_start=161
  _globals['_SENDCONNECTIONREQUEST']._serialized_end=204
  _globals['_REQUESTACCEPTED']._serialized_start=206
  _globals['_REQUESTACCEPTED']._serialized_end=241
  _globals['_WAITEXPERIMENTREQUEST']._serialized_start=243
  _globals['_WAITEXPERIMENTREQUEST']._serialized_end=293
  _globals['_WAITEXPERIMENTRESPONSE']._serialized_start=295
  _globals['_WAITEXPERIMENTRESPONSE']._serialized_end=344
  _globals['_GETEXPERIMENTDATAREQUEST']._serialized_start=346
  _globals['_GETEXPERIMENTDATAREQUEST']._serialized_end=424
  _globals['_EXPERIMENTDATA']._serialized_start=426
  _globals['_EXPERIMENTDATA']._serialized_end=474
  _globals['_UPDATEENVOYSTATUSREQUEST']._serialized_start=476
  _globals['_UPDATEENVOYSTATUSREQUEST']._serialized_end=547
  _globals['_UPDATEENVOYSTATUSRESPONSE']._serialized_start=549
  _globals['_UPDATEENVOYSTATUSRESPONSE']._serialized_end=632
  _globals['_EXPERIMENTINFO']._serialized_start=635
  _globals['_EXPERIMENTINFO']._serialized_end=773
  _globals['_SETNEWEXPERIMENTRESPONSE']._serialized_start=775
  _globals['_SETNEWEXPERIMENTRESPONSE']._serialized_end=817
  _globals['_ENVOYINFO']._serialized_start=820
  _globals['_ENVOYINFO']._serialized_end=1027
  _globals['_GETENVOYSREQUEST']._serialized_start=1029
  _globals['_GETENVOYSREQUEST']._serialized_end=1047
  _globals['_GETENVOYSRESPONSE']._serialized_start=1049
  _globals['_GETENVOYSRESPONSE']._serialized_end=1139
  _globals['_GETFLOWSTATEREQUEST']._serialized_start=1141
  _globals['_GETFLOWSTATEREQUEST']._serialized_end=1162
  _globals['_GETFLOWSTATERESPONSE']._serialized_start=1164
  _globals['_GETFLOWSTATERESPONSE']._serialized_end=1225
  _globals['_SENDRUNTIMEREQUEST']._serialized_start=1227
  _globals['_SENDRUNTIMEREQUEST']._serialized_end=1247
  _globals['_RUNTIMEREQUESTRESPONSE']._serialized_start=1249
  _globals['_RUNTIMEREQUESTRESPONSE']._serialized_end=1291
  _globals['_GETEXPERIMENTSTDOUTREQUEST']._serialized_start=1293
  _globals['_GETEXPERIMENTSTDOUTREQUEST']._serialized_end=1346
  _globals['_GETEXPERIMENTSTDOUTRESPONSE']._serialized_start=1348
  _globals['_GETEXPERIMENTSTDOUTRESPONSE']._serialized_end=1441
  _globals['_DIRECTOR']._serialized_start=1444
  _globals['_DIRECTOR']._serialized_end=2770
# @@protoc_insertion_point(module_scope)



# Copyright 2020-2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""
This file contains utility functions for Secure Aggregation Shamir's secret
sharing related operations.
"""

from Crypto.Protocol.SecretSharing import Shamir  # nosec B413

# The length of each chunk of secret used for creation of shares.
SECRET_CHUNK_LENGTH = 16


def create_secret_shares(
    secret: bytes,
    count: int,
    threshold: int,
) -> dict:
    """
    Using Shamir's secret sharing protocol, creates shares for the secret.
    pycryptodome's Shamir.split requires the secret to be of length
    SECRET_CHUNK_LENGTH. Since the secret length cannot always be set to be
    SECRET_CHUNK_LENGTH, we split the secret byte string into multiple byte
    string each of length SECRET_CHUNK_LENGTH and create shares for each of
    the SECRET_CHUNK_LENGTH chunks.
    During reconstruction, the SECRET_CHUNK_LENGTH chunks are recreated using
    the respective shares and then concatenated to form the original secret.
    Thus each of the collaborator indices have multiple shares (one for each
    chunk) for a single secret.

    Args:
        secret (bytes): Secret for which shares are to be created.
        count (int): Number of shares to be created for the secret.
        threshold (int): Minimum threshold of shares required for
            reconstruction of the secret.

    Returns:
        dict: Contains the mapping of the index to the shares that belong to
            the respective index.
    """
    shares = {}

    from Crypto.Util.Padding import pad  # nosec B413

    # Pad the secret to create a byte string of a length which is a multiple
    # of SECRET_CHUNK_LENGTH.
    secret = pad(secret, SECRET_CHUNK_LENGTH)
    # Divide the secret into multiple chunks.
    secret_chunks = [
        secret[i : i + SECRET_CHUNK_LENGTH] for i in range(0, len(secret), SECRET_CHUNK_LENGTH)
    ]
    # Create shares for each of the chunk.
    for chunk in secret_chunks:
        chunk_shares = Shamir.split(threshold, count, chunk)
        # Map the respective chunk share to the id it belongs to.
        for id, share in chunk_shares:
            if id not in shares:
                shares[id] = []
            shares[id].append(share.hex())

    shares = {k: ".".join(v) for k, v in shares.items()}

    return shares


def reconstruct_secret(shares: dict) -> bytes:
    """
    Args:
        shares (dict): Contains the mapping of the index to the chunk shares
            that belong to the respective index.

    Returns:
        bytes: Secret reconstructed from the chunk shares.
    """
    secret = b""

    shares = {k: v.split(".") for k, v in shares.items()}

    total_chunks = max(len(share) for share in shares.values())
    for chunk_index in range(total_chunks):
        # Create a list for the respective chunk with all the shares.
        chunk_shares = [(key, bytes.fromhex(shares[key][chunk_index])) for key in shares]
        # Reconstruct the chunk of the secret.
        secret_chunk = Shamir.combine(chunk_shares)
        # Concatenate the chunk to the secret.
        secret += secret_chunk

    from Crypto.Util.Padding import unpad  # nosec B413

    # Remove the padding from the secret.
    secret = unpad(secret, SECRET_CHUNK_LENGTH)

    return secret



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""KerasDataLoader module."""

import numpy as np

from openfl.federated.data.loader import DataLoader


class KerasDataLoader(DataLoader):
    """A class used to represent a Federation Data Loader for Keras models.

    Attributes:
        batch_size (int): Size of batches used for all data loaders.
        X_train (np.array): Training features.
        y_train (np.array): Training labels.
        X_valid (np.array): Validation features.
        y_valid (np.array): Validation labels.
    """

    def __init__(self, batch_size, **kwargs):
        """Initializes the KerasDataLoader object.

        Args:
            batch_size (int): The size of batches used for all data loaders.
            kwargs: Additional arguments to pass to the function.
        """
        self.batch_size = batch_size
        self.X_train = None
        self.y_train = None
        self.X_valid = None
        self.y_valid = None

        # Child classes should have init signature:
        # (self, batch_size, **kwargs), should call this __init__ and then
        # define self.X_train, self.y_train, self.X_valid, and self.y_valid

    def get_feature_shape(self):
        """Returns the shape of an example feature array.

        Returns:
            tuple: The shape of an example feature array.
        """
        return self.X_train[0].shape

    def get_train_loader(self, batch_size=None, num_batches=None):
        """Returns the data loader for the training data.

        Args:
            batch_size (int, optional): The batch size for the data loader
                (default is None).
            num_batches (int, optional): The number of batches for the data
                loader (default is None).

        Returns:
            DataLoader: The DataLoader object for the training data.
        """
        return self._get_batch_generator(
            X=self.X_train,
            y=self.y_train,
            batch_size=batch_size,
            num_batches=num_batches,
        )

    def get_valid_loader(self, batch_size=None):
        """Returns the data loader for the validation data.

        Args:
            batch_size (int, optional): The batch size for the data loader
                (default is None).

        Returns:
            DataLoader: The DataLoader object for the validation data.
        """
        return self._get_batch_generator(X=self.X_valid, y=self.y_valid, batch_size=batch_size)

    def get_train_data_size(self):
        """Returns the total number of training samples.

        Returns:
            int: The total number of training samples.
        """
        return self.X_train.shape[0]

    def get_valid_data_size(self):
        """Returns the total number of validation samples.

        Returns:
            int: The total number of validation samples.
        """
        return self.X_valid.shape[0]

    @staticmethod
    def _batch_generator(X, y, idxs, batch_size, num_batches):
        """Generates batches of data.

        Args:
            X (np.array): The input data.
            y (np.array): The label data.
            idxs (np.array): The index of the dataset.
            batch_size (int): The batch size for the data loader.
            num_batches (int): The number of batches.

        Yields:
            tuple: The input data and label data for each batch.
        """
        for i in range(num_batches):
            a = i * batch_size
            b = a + batch_size
            yield X[idxs[a:b]], y[idxs[a:b]]

    def _get_batch_generator(self, X, y, batch_size, num_batches=None):
        """Returns the dataset generator.

        Args:
            X (np.array): The input data.
            y (np.array): The label data.
            batch_size (int): The batch size for the data loader.
            num_batches (int, optional): The number of batches (default is
                None).

        Returns:
            generator: The dataset generator.
        """
        if batch_size is None:
            batch_size = self.batch_size

        # shuffle data indices
        idxs = np.random.permutation(np.arange(X.shape[0]))

        if num_batches is None:
            # compute the number of batches
            num_batches = int(np.ceil(X.shape[0] / batch_size))

        # build the generator and return it
        return self._batch_generator(X, y, idxs, batch_size, num_batches)



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Assigner module."""

import logging
from functools import wraps

logger = logging.getLogger(__name__)


class Assigner:
    r"""
    The task assigner maintains a list of tasks.

    Also it decides the policy for which collaborator should run those tasks.
    There may be many types of policies implemented, but a natural place to start
    is with a:

        - RandomGroupedTaskAssigner :
            Given a set of task groups, and a percentage,
            assign that task group to that percentage of collaborators in the federation.
            After assigning the tasks to collaborator, those tasks should be carried
            out each round (no reassignment between rounds).
        - GroupedTaskAssigner :
            Given task groups and a list of collaborators that belong to that task group,
            carry out tasks for each round of experiment.

    Attributes:
        tasks* (list of object): List of tasks to assign.
        authorized_cols (list of str): Collaborators.
        rounds (int): Number of rounds to train.
        all_tasks_in_groups (list): All tasks in groups.
        task_group_collaborators (dict): Task group collaborators.
        collaborators_for_task (dict): Collaborators for each task.
        collaborator_tasks (dict): Tasks for each collaborator.

    .. note::
        \* - ``tasks`` argument is taken from ``tasks`` section of FL plan YAML file.
    """

    def __init__(
        self,
        tasks,
        authorized_cols,
        rounds_to_train,
        selected_task_group: str = None,
        **kwargs,
    ):
        """Initializes the Assigner.

        Args:
            tasks (list of object): List of tasks to assign.
            authorized_cols (list of str): Collaborators.
            rounds_to_train (int): Number of training rounds.
            selected_task_group (str, optional): Selected task_group.
            **kwargs: Additional keyword arguments.
        """
        self.tasks = tasks
        self.authorized_cols = authorized_cols
        self.rounds = rounds_to_train
        self.selected_task_group = selected_task_group
        self.all_tasks_in_groups = []

        self.task_group_collaborators = {}
        self.collaborators_for_task = {}
        self.collaborator_tasks = {}

        self.define_task_assignments()

    def define_task_assignments(self):
        """Abstract method."""
        raise NotImplementedError

    def get_tasks_for_collaborator(self, collaborator_name, round_number):
        """Abstract method."""
        raise NotImplementedError

    def get_collaborators_for_task(self, task_name, round_number):
        """Abstract method."""
        raise NotImplementedError

    def is_task_group_evaluation(self):
        """Check if the selected task group is for 'evaluation' run.

        Returns:
            bool: True if the selected task group is 'evaluation', False otherwise.
        """
        if hasattr(self, "selected_task_group"):
            return self.selected_task_group == "evaluation"
        return False

    def get_all_tasks_for_round(self, round_number):
        """Return tasks for the current round.

        Currently all tasks are performed on each round,
        But there may be a reason to change this.

        Args:
            round_number (int): Round number.

        Returns:
            list: List of tasks for the current round.
        """
        return self.all_tasks_in_groups

    def get_aggregation_type_for_task(self, task_name):
        """Extract aggregation type from self.tasks.

        Args:
            task_name (str): Name of the task.

        Returns:
            str: Aggregation type for the task.
        """
        if "aggregation_type" not in self.tasks[task_name]:
            return None
        return self.tasks[task_name]["aggregation_type"]

    @classmethod
    def with_selected_task_group(cls, func):
        """Decorator to filter task groups based on selected_task_group.

        This decorator should be applied to define_task_assignments() method
        in Assigner subclasses to handle task_group filtering.
        """

        @wraps(func)
        def wrapper(self, *args, **kwargs):
            # Check if selection of task_group is applicable
            if hasattr(self, "selected_task_group") and self.selected_task_group is not None:
                # Verify task_groups exists before attempting filtering
                if not hasattr(self, "task_groups"):
                    logger.warning(
                        "Task group specified for selection but no task_groups found. "
                        "Skipping filtering. This might be intentional for custom assigners."
                    )
                    return func(self, *args, **kwargs)

                assert self.task_groups, "No task_groups defined in assigner."

                # Perform the filtering
                selected_task_groups = [
                    group for group in self.task_groups if group["name"] == self.selected_task_group
                ]

                assert len(selected_task_groups) == 1, (
                    f"Only one task group with name {self.selected_task_group} should exist"
                )

                # Since we have filtered to one of the task_groups, we need to ensure that
                # the selected_task_group percentage compute allocation is defaulted to 1.0
                current_percentage = selected_task_groups[0]["percentage"]
                logger.info(
                    f"`percentage` for task_group {self.selected_task_group} is "
                    f"{current_percentage}, setting it to 1.0"
                )
                selected_task_groups[0]["percentage"] = 1.0

                self.task_groups = selected_task_groups

            # Call the original method
            return func(self, *args, **kwargs)

        return wrapper



# Copyright (C) 2020-2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""You may copy this file as the starting point of your own model."""
import os
import numpy as np
from tqdm import tqdm
import argparse


def create_data(collaborators, dst, total_dataset_size_per_col):
    num_samples = 1
    input_shape = (160, 160, 128)
    X_train = np.random.random((num_samples,) + input_shape)
    y_train = np.random.random((num_samples,) + input_shape)

    # Save the arrays
    os.makedirs(dst + "/1", exist_ok=True)
    np.save(dst + "/1/X_train.npy", X_train)
    np.save(dst + "/1/y_train.npy", y_train)
    single_file_size_MB = os.path.getsize(dst + "/1/X_train.npy") / 1024 / 1024
    single_file_size_MB += os.path.getsize(dst + "/1/y_train.npy") / 1024 / 1024
    required_file_size_MB = total_dataset_size_per_col * collaborators
    files_to_make = required_file_size_MB // single_file_size_MB
    files_per_collaborator = int(files_to_make // collaborators)
    for col in range(1, collaborators + 1):
        print("making files for collaborator", col)
        os.makedirs(dst + f"/{col}", exist_ok=True)
        for i in tqdm(range(files_per_collaborator)):
            X_train = np.random.random((num_samples,) + input_shape)
            y_train = np.random.random((num_samples,) + input_shape)
            np.save(dst + f"/{col}/X_train_{i}.npy", X_train)
            np.save(dst + f"/{col}/y_train_{i}.npy", y_train)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--num_collaborators", type=int, help="Number of collaborators")
    parser.add_argument(
        "--total_dataset_size_per_col_MB", type=int, help="Total dataset size per collaborator"
    )
    args = parser.parse_args()

    num_collaborators = args.num_collaborators
    total_dataset_size_per_col = args.total_dataset_size_per_col_MB

    dst = "data"
    create_data(num_collaborators, dst, total_dataset_size_per_col)



import importlib

def get_interop_server(framework: str = 'Flower') -> object:
    if framework == 'Flower':
        try:
            module = importlib.import_module('src.grpc.connector.flower.interop_server')
            return module.FlowerInteropServer
        except ImportError:
            print("Flower is not installed.")
            return None



# Copyright (C) 2020-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""

from logging import getLogger

import numpy as np
from torchvision import datasets
from torchvision import transforms

logger = getLogger(__name__)


def one_hot(labels, classes):
    """
    One Hot encode a vector.

    Args:
        labels (list):  List of labels to onehot encode
        classes (int): Total number of categorical classes

    Returns:
        np.array: Matrix of one-hot encoded labels
    """
    return np.eye(classes)[labels]


def _load_raw_datashards(shard_num, collaborator_count, transform=None):
    """
    Load the raw data by shard.

    Returns tuples of the dataset shard divided into training and validation.

    Args:
        shard_num (int): The shard number to use
        collaborator_count (int): The number of collaborators in the federation
        transform: torchvision.transforms.Transform to apply to images

    Returns:
        2 tuples: (image, label) of the training, validation dataset
    """
    train_data, val_data = (
        datasets.MNIST('data', train=train, download=True, transform=transform)
        for train in (True, False)
    )
    X_train_tot, y_train_tot = train_data.train_data, train_data.train_labels
    X_valid_tot, y_valid_tot = val_data.test_data, val_data.test_labels

    # create the shards
    shard_num = int(shard_num)
    X_train = X_train_tot[shard_num::collaborator_count].unsqueeze(1).float()
    y_train = y_train_tot[shard_num::collaborator_count]

    X_valid = X_valid_tot[shard_num::collaborator_count].unsqueeze(1).float()
    y_valid = y_valid_tot[shard_num::collaborator_count]

    return (X_train, y_train), (X_valid, y_valid)


def load_mnist_shard(shard_num, collaborator_count,
                     categorical=False, channels_last=True, **kwargs):
    """
    Load the MNIST dataset.

    Args:
        shard_num (int): The shard to use from the dataset
        collaborator_count (int): The number of collaborators in the
                                  federation
        categorical (bool): True = convert the labels to one-hot encoded
                            vectors (Default = True)
        channels_last (bool): True = The input images have the channels
                              last (Default = True)
        **kwargs: Additional parameters to pass to the function

    Returns:
        list: The input shape
        int: The number of classes
        numpy.ndarray: The training data
        numpy.ndarray: The training labels
        numpy.ndarray: The validation data
        numpy.ndarray: The validation labels
    """
    num_classes = 10

    (X_train, y_train), (X_valid, y_valid) = _load_raw_datashards(
        shard_num, collaborator_count, transform=transforms.ToTensor())

    logger.info(f'MNIST > X_train Shape : {X_train.shape}')
    logger.info(f'MNIST > y_train Shape : {y_train.shape}')
    logger.info(f'MNIST > Train Samples : {X_train.shape[0]}')
    logger.info(f'MNIST > Valid Samples : {X_valid.shape[0]}')

    if categorical:
        # convert class vectors to binary class matrices
        y_train = one_hot(y_train, num_classes)
        y_valid = one_hot(y_valid, num_classes)

    return num_classes, X_train, y_train, X_valid, y_valid



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""

from openfl.federated import KerasDataLoader
from .brats_utils import load_from_nifti


class KerasBratsInMemory(KerasDataLoader):
    """Keras Data Loader for the BraTS dataset."""

    def __init__(self, data_path, batch_size, percent_train=0.8, pre_split_shuffle=True, **kwargs):
        """Initialize.

        Args:
            data_path: The file path for the BraTS dataset
            batch_size (int): The batch size to use
            percent_train (float): The percentage of the data to use for training (Default=0.8)
            pre_split_shuffle (bool): True= shuffle the dataset before
            performing the train/validate split (Default=True)
            **kwargs: Additional arguments, passed to super init and load_from_nifti

        Returns:
            Data loader with BraTS data
        """
        super().__init__(batch_size, **kwargs)

        X_train, y_train, X_valid, y_valid = load_from_nifti(parent_dir=data_path,
                                                             percent_train=percent_train,
                                                             shuffle=pre_split_shuffle,
                                                             **kwargs)
        self.X_train = X_train
        self.y_train = y_train
        self.X_valid = X_valid
        self.y_valid = y_valid



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Plan module."""

import inspect
import os
from hashlib import sha384
from importlib import import_module, reload
from logging import getLogger
from os.path import splitext
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple

from yaml import SafeDumper, dump, safe_load

from openfl.experimental.workflow.interface.cli.cli_helper import WORKSPACE
from openfl.experimental.workflow.transport import AggregatorGRPCClient, AggregatorGRPCServer
from openfl.utilities.utils import getfqdn_env

if TYPE_CHECKING:
    from openfl.experimental.workflow.component import Aggregator, Collaborator
    from openfl.experimental.workflow.interface import FLSpec

SETTINGS = "settings"
TEMPLATE = "template"
DEFAULTS = "defaults"
AUTO = "auto"


class Plan:
    """A class used to represent a Federated Learning plan.

    This class provides methods to manage and manipulate federated learning
    plans.

    Attributes:
        logger (Logger): Logger instance for the class.
        config (dict): Dictionary containing patched plan definition.
        authorized_cols (list): Authorized collaborator list.
        cols_data_paths (dict): Collaborator data paths dictionary.
        collaborator_ (Collaborator): Collaborator object.
        aggregator_ (Aggregator): Aggregator object.
        server_ (AggregatorGRPCServer): gRPC server object.
        client_ (AggregatorGRPCClient): gRPC client object.
        hash_ (str): Hash of the instance.
    """

    logger = getLogger(__name__)

    @staticmethod
    def load(yaml_path: Path, default: dict = None) -> dict:
        """Load the plan from YAML file.

        Args:
            yaml_path (Path): Path to the YAML file.
            default (dict, optional): Default plan configuration.
                Defaults to {}.

        Returns:
            dict: Plan configuration loaded from the YAML file.
        """
        if default is None:
            default = {}
        if yaml_path and yaml_path.exists():
            return safe_load(yaml_path.read_text())
        return default

    @staticmethod
    def dump(yaml_path, config, freeze=False) -> None:
        """Dump the plan config to YAML file.

        Args:
            yaml_path (Path): Path to the YAML file.
            config (dict): Plan configuration to be dumped.
            freeze (bool, optional): Flag to freeze the plan. Defaults to
                False.
        """

        class NoAliasDumper(SafeDumper):
            def ignore_aliases(self, data):
                return True

        if freeze:
            plan = Plan()
            plan.config = config
            frozen_yaml_path = Path(f"{yaml_path.parent}/{yaml_path.stem}_{plan.hash[:8]}.yaml")
            if frozen_yaml_path.exists():
                Plan.logger.info(f"{yaml_path.name} is already frozen")
                return
            frozen_yaml_path.write_text(dump(config))
            frozen_yaml_path.chmod(0o400)
            Plan.logger.info(f"{yaml_path.name} frozen successfully")
        else:
            yaml_path.write_text(dump(config))

    @staticmethod
    def parse(
        plan_config_path: Path,
        cols_config_path: Path = None,
        data_config_path: Path = None,
        resolve=True,
    ) -> "Plan":
        """Parse the Federated Learning plan.

        Args:
            plan_config_path (string): The filepath to the federated learning
                                       plan
            cols_config_path (string): The filepath to the federation
                                       collaborator list [optional]
            data_config_path (string): The filepath to the federation
                                       collaborator data configuration
                                       [optional]
        Returns:
            A federated learning plan object
        """
        try:
            plan = Plan()
            plan.config = Plan.load(plan_config_path)  # load plan configuration
            plan.name = plan_config_path.name
            plan.files = [plan_config_path]  # collect all the plan files

            # ensure 'settings' appears in each top-level section
            for section in plan.config.keys():
                if plan.config[section].get(SETTINGS) is None:
                    plan.config[section][SETTINGS] = {}

            # walk the top level keys and load 'defaults' in sorted order
            for section in sorted(plan.config.keys()):
                defaults = plan.config[section].pop(DEFAULTS, None)

                if defaults is not None:
                    defaults = WORKSPACE / "workspace" / defaults

                    plan.files.append(defaults)

                    if resolve:
                        Plan.logger.info(
                            f"Loading DEFAULTS for section [red]{section}[/] "
                            f"from file [red]{defaults}[/].",
                            extra={"markup": True},
                        )

                    defaults = Plan.load(Path(defaults))

                    if SETTINGS in defaults:
                        # override defaults with section settings
                        defaults[SETTINGS].update(plan.config[section][SETTINGS])
                        plan.config[section][SETTINGS] = defaults[SETTINGS]

                    defaults.update(plan.config[section])

                    plan.config[section] = defaults

            plan.authorized_cols = Plan.load(cols_config_path).get("collaborators", [])

            if resolve:
                plan.resolve()

                Plan.logger.info(
                    f"Parsing Federated Learning Plan : [green]SUCCESS[/] : "
                    f"[blue]{plan_config_path}[/].",
                    extra={"markup": True},
                )
                Plan.dump(plan_config_path, plan.config)
                Plan.logger.info(dump(plan.config))

            return plan

        except Exception:
            Plan.logger.exception(
                f"Parsing Federated Learning Plan : [red]FAILURE[/] : [blue]{plan_config_path}[/].",
                extra={"markup": True},
            )
            raise

    @staticmethod
    def accept_args(cls) -> bool:
        """Determines whether a class's constructor (__init__ method) accepts
        variable positional arguments (*args).

        Returns:
            Boolean: True or False
        """
        init_signature = inspect.signature(cls.__init__)
        for param in init_signature.parameters.values():
            if param.kind == param.VAR_POSITIONAL:
                return True
        return False

    @staticmethod
    def build(template, settings, **override) -> object:
        """Create an instance of a openfl Component or Federated
        DataLoader/TaskRunner.

        Args:
            template: Fully qualified class template path
            settings: Keyword arguments to class constructor

        Returns:
            A Python object
        """
        class_name = splitext(template)[1].strip(".")
        module_path = splitext(template)[0]

        Plan.logger.info(f"Building `{template}` Module.")
        Plan.logger.debug(f"Settings {settings}")
        Plan.logger.debug(f"Override {override}")

        settings.update(**override)
        module = import_module(module_path)
        # Reload the module to ensure the flow is rebuilt with updated changes
        module = reload(module)

        if Plan.accept_args(getattr(module, class_name)):
            args = list(settings.values())
            instance = getattr(module, class_name)(*args)
        else:
            instance = getattr(module, class_name)(**settings)

        return instance

    @staticmethod
    def import_(template) -> object:
        """Import an instance of a openfl Component or Federated
        DataLoader/TaskRunner.

        Args:
            template: Fully qualified object path

        Returns:
            A Python object
        """
        class_name = splitext(template)[1].strip(".")
        module_path = splitext(template)[0]
        Plan.logger.info(
            f"Importing [red]🡆[/] Object [red]{class_name}[/] from [red]{module_path}[/] Module.",
            extra={"markup": True},
        )
        module = import_module(module_path)
        # FIXME: Reload module to refresh private attributes; optimize this.
        module = reload(module)
        instance = getattr(module, class_name)

        return instance

    def __init__(self) -> None:
        """Initialize."""
        self.config = {}  # dictionary containing patched plan definition
        self.authorized_cols = []  # authorized collaborator list
        self.cols_data_paths = {}  # collaborator data paths dict

        self.collaborator_ = None  # collaborator object
        self.aggregator_ = None  # aggregator object

        self.server_ = None  # gRPC server object
        self.client_ = None  # gRPC client object

        self.hash_ = None

    @property
    def hash(self) -> str:  # NOQA
        """Generate hash for this instance."""
        self.hash_ = sha384(dump(self.config).encode("utf-8"))
        Plan.logger.info(
            f"FL-Plan hash is [blue]{self.hash_.hexdigest()}[/]",
            extra={"markup": True},
        )

        return self.hash_.hexdigest()

    def resolve(self) -> None:
        """Resolve the federation settings."""
        self.federation_uuid = f"{self.name}_{self.hash[:8]}"
        self.aggregator_uuid = f"aggregator_{self.federation_uuid}"

        self.rounds_to_train = self.config["aggregator"][SETTINGS]["rounds_to_train"]

        if self.config["network"][SETTINGS]["agg_addr"] == AUTO:
            self.config["network"][SETTINGS]["agg_addr"] = getfqdn_env()

        if self.config["network"][SETTINGS]["agg_port"] == AUTO:
            self.config["network"][SETTINGS]["agg_port"] = (
                int(self.hash[:8], 16) % (60999 - 49152) + 49152
            )

    def get_aggregator(self, director_config=None) -> "Aggregator":
        """Get federation aggregator.

        Args:
            director_config: Path to director config file.
                Defaults to None

        Returns:
            self.aggregator_ (Aggregator): The federation aggregator.
        """
        defaults = self.config.get(
            "aggregator",
            {TEMPLATE: "openfl.experimental.workflow.Aggregator", SETTINGS: {}},
        )

        defaults[SETTINGS]["aggregator_uuid"] = self.aggregator_uuid
        defaults[SETTINGS]["federation_uuid"] = self.federation_uuid
        defaults[SETTINGS]["authorized_cols"] = self.authorized_cols

        private_attrs_callable, private_attrs_kwargs, private_attributes = self.get_private_attr(
            "aggregator", director_config
        )
        defaults[SETTINGS]["private_attributes_callable"] = private_attrs_callable
        defaults[SETTINGS]["private_attributes_kwargs"] = private_attrs_kwargs
        defaults[SETTINGS]["private_attributes"] = private_attributes

        defaults[SETTINGS]["flow"] = self.get_flow()
        defaults[SETTINGS]["checkpoint"] = defaults[SETTINGS]["flow"].checkpoint

        log_metric_callback = defaults[SETTINGS].get("log_metric_callback")
        if log_metric_callback:
            if isinstance(log_metric_callback, dict):
                log_metric_callback = Plan.import_(**log_metric_callback)
            elif not callable(log_metric_callback):
                raise TypeError(
                    f"log_metric_callback should be callable object "
                    f"or be import from code part, get {log_metric_callback}"
                )
        defaults[SETTINGS]["log_metric_callback"] = log_metric_callback

        if self.aggregator_ is None:
            self.aggregator_ = Plan.build(**defaults)

        return self.aggregator_

    def get_collaborator(
        self,
        collaborator_name,
        root_certificate=None,
        private_key=None,
        certificate=None,
        client=None,
        tls=True,
        envoy_config=None,
    ) -> "Collaborator":
        """Get collaborator.

        This method retrieves a collaborator. If the collaborator does not
        exist, it is built using the configuration settings and the provided
        parameters.

        Args:
            collaborator_name (str): Name of the collaborator.
            root_certificate (str, optional): Root certificate for the
                collaborator. Defaults to None.
            private_key (str, optional): Private key for the collaborator.
                Defaults to None.
            certificate (str, optional): Certificate for the collaborator.
                Defaults to None.
            client (Client, optional): Client for the collaborator. Defaults
                to None.
            tls (bool): Whether to use TLS for the connection.
            envoy_config (Path): Path to envoy_config.yaml. Defaults
                to None.

        Returns:
            self.collaborator_ (Collaborator): The collaborator instance.
        """
        defaults = self.config.get(
            "collaborator",
            {TEMPLATE: "openfl.experimental.workflow.Collaborator", SETTINGS: {}},
        )

        defaults[SETTINGS]["collaborator_name"] = collaborator_name
        defaults[SETTINGS]["aggregator_uuid"] = self.aggregator_uuid
        defaults[SETTINGS]["federation_uuid"] = self.federation_uuid

        private_attrs_callable, private_attrs_kwargs, private_attributes = self.get_private_attr(
            collaborator_name, envoy_config
        )
        defaults[SETTINGS]["private_attributes_callable"] = private_attrs_callable
        defaults[SETTINGS]["private_attributes_kwargs"] = private_attrs_kwargs
        defaults[SETTINGS]["private_attributes"] = private_attributes

        if client is not None:
            defaults[SETTINGS]["client"] = client
        else:
            defaults[SETTINGS]["client"] = self.get_client(
                collaborator_name,
                self.aggregator_uuid,
                self.federation_uuid,
                root_certificate,
                private_key,
                certificate,
                tls,
            )

        if self.collaborator_ is None:
            self.collaborator_ = Plan.build(**defaults)

        return self.collaborator_

    def get_client(
        self,
        collaborator_name,
        aggregator_uuid,
        federation_uuid,
        root_certificate=None,
        private_key=None,
        certificate=None,
        tls=True,
    ) -> AggregatorGRPCClient:
        """Get gRPC client for the specified collaborator.

        Args:
            collaborator_name (str): Name of the collaborator.
            aggregator_uuid (str): UUID of the aggregator.
            federation_uuid (str): UUID of the federation.
            root_certificate (str, optional): Root certificate for the
                collaborator. Defaults to None.
            private_key (str, optional): Private key for the collaborator.
                Defaults to None.
            certificate (str, optional): Certificate for the collaborator.
                Defaults to None.
            tls (bool): Whether to use TLS for the connection.

        Returns:
            AggregatorGRPCClient: gRPC client for the specified collaborator.
        """
        if tls and not (root_certificate and private_key and certificate):
            common_name = collaborator_name
            root_certificate = "cert/cert_chain.crt"
            certificate = f"cert/client/col_{common_name}.crt"
            private_key = f"cert/client/col_{common_name}.key"

        client_args = self.config["network"][SETTINGS]

        # patch certificates

        client_args["root_certificate"] = root_certificate
        client_args["certificate"] = certificate
        client_args["private_key"] = private_key

        client_args["aggregator_uuid"] = aggregator_uuid
        client_args["federation_uuid"] = federation_uuid

        if self.client_ is None:
            self.client_ = AggregatorGRPCClient(**client_args)

        return self.client_

    def get_server(
        self,
        root_certificate=None,
        private_key=None,
        certificate=None,
        tls=True,
        director_config=None,
        **kwargs,
    ) -> AggregatorGRPCServer:
        """Get gRPC server of the aggregator instance.

        Args:
            root_certificate (str, optional): Root certificate for the server.
                Defaults to None.
            private_key (str, optional): Private key for the server. Defaults
                to None.
            certificate (str, optional): Certificate for the server. Defaults
                to None.
            tls (bool): Whether to use TLS for the connection.
            director_config (Path): Path to director_config.yaml. Defaults
                to None.
            **kwargs: Additional keyword arguments.

        Returns:
            AggregatorGRPCServer: gRPC server of the aggregator instance.
        """
        if tls and not (root_certificate and private_key and certificate):
            common_name = self.config["network"][SETTINGS]["agg_addr"].lower()
            root_certificate = "cert/cert_chain.crt"
            certificate = f"cert/server/agg_{common_name}.crt"
            private_key = f"cert/server/agg_{common_name}.key"

        server_args = self.config["network"][SETTINGS]

        # patch certificates

        server_args.update(kwargs)
        server_args["root_certificate"] = root_certificate
        server_args["certificate"] = certificate
        server_args["private_key"] = private_key
        server_args["tls"] = tls

        server_args["aggregator"] = self.get_aggregator(director_config)

        if self.server_ is None:
            self.server_ = AggregatorGRPCServer(**server_args)

        return self.server_

    def get_flow(self) -> "FLSpec":
        """Instantiates federated flow object.

        Returns:
            flow_: FLSpec instance
        """
        defaults = self.config.get(
            "federated_flow",
            {TEMPLATE: self.config["federated_flow"]["template"], SETTINGS: {}},
        )
        defaults = self.import_kwargs_modules(defaults)

        self.flow_ = Plan.build(**defaults)
        return self.flow_

    def import_kwargs_modules(self, defaults) -> Dict[str, Any]:
        """
        Imports and resolves class references in a nested settings structure.
        Args:
            defaults (Dict[str, Any]): A dictionary of settings, containing module paths
                and class names as strings in its nested structure.

        Returns:
            Dict[str, Any]: The updated settings dictionary with resolved classes and attributes
                from the imported modules.
        """

        def import_nested_settings(settings):
            for key, value in settings.items():
                if isinstance(value, dict):
                    settings[key] = import_nested_settings(value)
                elif isinstance(value, str):
                    class_name = splitext(value)[1].strip(".")
                    if class_name:
                        module_path = splitext(value)[0]
                        try:
                            if import_module(module_path):
                                module = import_module(module_path)
                                value_defaults_data = {
                                    "template": value,
                                    "settings": settings.get("settings", {}),
                                }
                                attr = getattr(module, class_name)

                                if not inspect.isclass(attr):
                                    settings[key] = attr
                                else:
                                    settings = Plan.build(**value_defaults_data)
                        except ImportError:
                            raise ImportError(f"Cannot import {value}.")
            return settings

        defaults[SETTINGS] = import_nested_settings(defaults[SETTINGS])
        return defaults

    def get_private_attr(
        self, private_attr_name=None, config=None
    ) -> Tuple[Optional[dict], Optional[dict], dict]:
        """
        Retrieves private attributes defined in a configuration or data file.

        Args:
            private_attr_name (str): The name of the participant (Aggregator or Collaborator)
                whose private attribute is to be retrieved.
            config (Path): Path to the config file.

        Returns:
            Tuple: A tuple containing:
                - private_attrs_callable (Optional[dict]): A dictionary containing callable
                    function information, or None if not applicable.
                - private_attrs_kwargs (Optional[dict]): A dictionary of arguments for the
                    callable function, or None if not applicable.
                - private_attributes (dict): A dictionary of private attributes,
                    or an empty dictionary if none are found.
        """
        private_attrs_callable = private_attrs_kwargs = None
        private_attributes = {}

        data_yaml = "plan/data.yaml"

        if config or (os.path.exists(data_yaml) and os.path.isfile(data_yaml)):
            d = Plan.load(config) if config else Plan.load(Path(data_yaml).absolute())

            if d and d.get(private_attr_name, None):
                callable_func = d.get(private_attr_name, {}).get("callable_func")
                private_attributes = d.get(private_attr_name, {}).get("private_attributes")
                if callable_func and private_attributes:
                    logger = getLogger(__name__)
                    logger.warning(
                        f"Warning: {private_attr_name} private attributes "
                        "will be initialized via callable and "
                        "attributes directly specified "
                        "will be ignored"
                    )

                if callable_func is not None:
                    private_attrs_callable = {
                        "template": d.get(private_attr_name)["callable_func"]["template"]
                    }

                    private_attrs_kwargs = self.import_kwargs_modules(
                        d.get(private_attr_name)["callable_func"]
                    )["settings"]

                    if isinstance(private_attrs_callable, dict):
                        private_attrs_callable = Plan.import_(**private_attrs_callable)
                elif private_attributes:
                    private_attributes = Plan.import_(
                        d.get(private_attr_name)["private_attributes"]
                    )
                elif not callable(private_attrs_callable):
                    raise TypeError(
                        f"private_attrs_callable should be callable object "
                        f"or be import from code part, get {private_attrs_callable}"
                    )

        return (
            private_attrs_callable,
            private_attrs_kwargs,
            private_attributes,
        )



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""openfl.experimental.workflow.runtime package LocalRuntime class."""

from __future__ import annotations

import gc
import importlib
import math
import os
from copy import deepcopy
from logging import getLogger
from typing import Any, Callable, Dict, List, Optional, Type

import ray

from openfl.experimental.workflow.interface.fl_spec import FLSpec
from openfl.experimental.workflow.interface.participants import Aggregator, Collaborator
from openfl.experimental.workflow.runtime.runtime import Runtime
from openfl.experimental.workflow.utilities import (
    ResourcesNotAvailableError,
    aggregator_to_collaborator,
    check_resource_allocation,
    checkpoint,
    filter_attributes,
    generate_artifacts,
    get_number_of_gpus,
)


class RayExecutor:
    """Class for executing tasks using the Ray framework."""

    def __init__(self):
        """Initializes the RayExecutor object."""
        self.__remote_contexts = []

    def ray_call_put(
        self,
        participant: Any,
        ctx: Any,
        f_name: str,
        callback: Callable,
        clones: Optional[Any] = None,
    ) -> None:
        """Execute f_name from inside participant (Aggregator or Collaborator)
        class with the context of clone (ctx).

        Args:
            participant (Any): The participant (Aggregator or Collaborator) to
                execute the function in.
            ctx (Any): The context to execute the function in.
            f_name (str): The name of the function to execute.
            callback (Callable): The callback to execute after the function.
            clones (Optional[Any], optional): The clones to use in the
                function. Defaults to None.
        """
        if clones is not None:
            self.__remote_contexts.append(
                participant.execute_func.remote(ctx, f_name, callback, clones)
            )
        else:
            self.__remote_contexts.append(participant.execute_func.remote(ctx, f_name, callback))

    def ray_call_get(self) -> List[Any]:
        """Get remote clones and delete ray references of clone (ctx) and,
        reclaim memory.

        Returns:
            List[Any]: The list of remote clones.
        """
        clones = ray.get(self.__remote_contexts)
        del self.__remote_contexts
        self.__remote_contexts = []

        return clones


def ray_group_assign(collaborators, num_actors=1):
    """Assigns collaborators to resource groups which share a CUDA context.

    Args:
        collaborators (list): The list of collaborators.
        num_actors (int, optional): Number of actors to distribute
            collaborators to. Defaults to 1.

    Returns:
        list: A list of GroupMember instances.
    """

    class GroupMember:
        """A utility class that manages the collaborator and its group.

        This class maintains compatibility with runtime execution by assigning
        attributes for each function in the Collaborator interface in
        conjunction with RemoteHelper.
        """

        def __init__(self, collaborator_actor, collaborator):
            """Initializes a new instance of the GroupMember class.

            Args:
                collaborator_actor: The collaborator actor.
                collaborator: The collaborator.
            """

            all_methods = [
                method for method in dir(Collaborator) if callable(getattr(Collaborator, method))
            ]
            external_methods = [method for method in all_methods if (method[0] != "_")]
            self.collaborator_actor = collaborator_actor
            self.collaborator = collaborator
            for method in external_methods:
                setattr(
                    self,
                    method,
                    RemoteHelper(self.collaborator_actor, self.collaborator, method),
                )

    class RemoteHelper:
        """A utility class to maintain compatibility with RayExecutor.

        This class returns a lambda function that uses
        collaborator_actor.execute_from_col to run a given function from the
        given collaborator.
        """

        # once ray_grouped replaces the current ray runtime this class can be
        # replaced with a function that returns the lambda function, using a
        # function is necessary because this is used in setting multiple
        # functions in a loop and lambda takes the reference to self.f_name and
        # not the value so we need to change scope to avoid self.f_name from
        # changing as the loop  progresses
        def __init__(self, collaborator_actor, collaborator, f_name) -> None:
            """Initializes a new instance of the RemoteHelper class.

            Args:
                collaborator_actor: The collaborator actor.
                collaborator: The collaborator.
                f_name (str): The name of the function.
            """
            self.f_name = f_name
            self.collaborator_actor = collaborator_actor
            self.collaborator = collaborator
            self.f = lambda *args, **kwargs: self.collaborator_actor.execute_from_col.remote(
                self.collaborator, self.f_name, *args, **kwargs
            )

        def remote(self, *args, **kwargs):
            """Executes the function with the given arguments and keyword
            arguments.

            Args:
                *args: The arguments to pass to the function.
                **kwargs: The keyword arguments to pass to the function.

            Returns:
                The result of the function execution.
            """
            return self.f(*args, *kwargs)

    collaborator_ray_refs = []
    collaborators_per_group = math.ceil(len(collaborators) / num_actors)
    times_called = 0
    # logic to sort collaborators by gpus, if collaborators have the same
    # number of gpu then they are sorted by cpu
    cpu_magnitude = len(str(abs(max([i.num_cpus for i in collaborators]))))
    min_gpu = min([i.num_gpus for i in collaborators])
    min_gpu = max(min_gpu, 0.0001)
    collaborators_sorted_by_gpucpu = sorted(
        collaborators,
        key=lambda x: x.num_gpus / min_gpu * 10**cpu_magnitude + x.num_cpus,
    )
    initializations = []
    collaborator_actor = None

    for collaborator in collaborators_sorted_by_gpucpu:
        # initialize actor group
        if times_called % collaborators_per_group == 0:
            max_num_cpus = max(
                [
                    i.num_cpus
                    for i in collaborators_sorted_by_gpucpu[
                        times_called : times_called + collaborators_per_group
                    ]
                ]
            )
            max_num_gpus = max(
                [
                    i.num_gpus
                    for i in collaborators_sorted_by_gpucpu[
                        times_called : times_called + collaborators_per_group
                    ]
                ]
            )
            print(f"creating actor with {max_num_cpus}, {max_num_gpus}")
            collaborator_actor = (
                ray.remote(RayGroup)
                .options(
                    num_cpus=max_num_cpus, num_gpus=max_num_gpus
                )  # max_concurrency=max_concurrency)
                .remote()
            )

        if collaborator_actor is None:
            raise ValueError("collaborator_actor has not been initialized.")

        # add collaborator to actor group
        initializations.append(collaborator_actor.append.remote(collaborator))

        times_called += 1

        # append GroupMember to output list
        collaborator_ray_refs.append(GroupMember(collaborator_actor, collaborator.get_name()))
    # Wait for all collaborators to be created on actors
    ray.get(initializations)

    return collaborator_ray_refs


class RayGroup:
    """A Ray actor that manages a group of collaborators.

    This class allows for the execution of functions from a specified
    collaborator using the execute_from_col method. The collaborators are
    stored in a dictionary where the key is the collaborator's name.
    """

    def __init__(self):
        """Initializes a new instance of the RayGroup class."""
        self.collaborators = {}

    def append(
        self,
        collaborator: Collaborator,
    ):
        """Appends a new collaborator to the group.

        Args:
            name (str): The name of the collaborator.
            private_attributes_callable (Callable): A callable that sets the
                private attributes of the collaborator.
            **kwargs: Additional keyword arguments.
        """

        if collaborator.private_attributes_callable is not None:
            self.collaborators[collaborator.name] = Collaborator(
                name=collaborator.name,
                private_attributes_callable=collaborator.private_attributes_callable,
                **collaborator.kwargs,
            )
        elif collaborator.private_attributes is not None:
            self.collaborators[collaborator.name] = Collaborator(
                name=collaborator.name,
                **collaborator.kwargs,
            )
            self.collaborators[collaborator.name].initialize_private_attributes(
                collaborator.private_attributes
            )

    def execute_from_col(self, name, internal_f_name, *args, **kwargs):
        """Executes a function from a specified collaborator.

        Args:
            name (str): The name of the collaborator.
            internal_f_name (str): The name of the function to execute.
            *args: Additional arguments to pass to the function.
            **kwargs: Additional keyword arguments to pass to the function.

        Returns:
            The result of the function execution.
        """
        f = getattr(self.collaborators[name], internal_f_name)
        return f(*args, **kwargs)

    def get_collaborator(self, name):
        """Retrieves a collaborator from the group by name.

        Args:
            name (str): The name of the collaborator.

        Returns:
            The collaborator instance.
        """
        return self.collaborators[name]


class LocalRuntime(Runtime):
    """Class for a local runtime, derived from the Runtime class.

    Attributes:
        aggregator (Type[Aggregator]): The aggregator participant.
        __collaborators (dict): The collaborators, stored as a dictionary of
            names to participants.
        backend (str): The backend that will execute the tasks.
    """

    def __init__(
        self,
        aggregator: Dict = None,
        collaborators: Dict = None,
        backend: str = "single_process",
        **kwargs,
    ) -> None:
        """Initializes the LocalRuntime object to run the flow on a single
        node, with an optional aggregator, an optional list of collaborators,
        an optional backend, and additional keyword arguments.

        Args:
            aggregator (Type[Aggregator], optional): The aggregator instance
                that holds private attributes.
            collaborators (List[Type[Collaborator]], optional): A list of
                collaborators; each with their own private attributes.
            backend (str, optional): The backend that will execute the tasks.
            Defaults to "single_process".
                Available options are:
                - 'single_process': (default) Executes every task within the
                  same process.
                - 'ray': Executes tasks using the Ray library. We use ray
                  actors called RayGroups to runs tasks in their own isolated
                  process. Each participant is distributed into a ray group.
                  The RayGroups run concurrently while participants in the
                  group run serially.
                  The default is 1 RayGroup and can be changed by using the
                  num_actors=1 kwarg. By using more RayGroups more concurrency
                  is allowed with the trade off being that each RayGroup has
                  extra memory overhead in the form of extra CUDA CONTEXTS.

                  Also the ray runtime supports GPU isolation using Ray's
                  'num_gpus' argument, which can be passed in through the
                  collaborator placement decorator.

        Raises:
            ValueError: If the provided backend value is not 'ray' or
                'single_process'.

        Example:
            @collaborator(num_gpus=1)
            def some_collaborator_task(self):
                # Task implementation
            ...

            By selecting num_gpus=1, the task is guaranteed exclusive GPU
            access. If the system has one GPU, collaborator tasks will run
            sequentially.
        """
        super().__init__()
        if backend not in ["ray", "single_process"]:
            raise ValueError(
                f"Invalid 'backend' value '{backend}', accepted values are "
                + "'ray', or 'single_process'"
            )
        if backend == "ray":
            if not ray.is_initialized():
                dh = kwargs.get("dashboard_host", "127.0.0.1")
                dp = kwargs.get("dashboard_port", 5252)
                ray.init(dashboard_host=dh, dashboard_port=dp)

            self.num_actors = kwargs.get("num_actors", 1)
        self.backend = backend
        self.logger = getLogger(__name__)
        if aggregator is not None:
            self.aggregator = self.__get_aggregator_object(aggregator)

        if collaborators is not None:
            self.collaborators = self.__get_collaborator_object(collaborators)

    def __get_aggregator_object(self, aggregator: Type[Aggregator]) -> Any:
        """Get aggregator object based on localruntime backend.

        If the backend is 'single_process', it returns the aggregator directly.
        If the backend is 'ray', it creates a Ray actor for the aggregator
        with the specified resources.

        Args:
            aggregator (Type[Aggregator]): The aggregator class to instantiate.

        Returns:
            Any: The aggregator object or a reference to the Ray actor
                representing the aggregator.

        Raises:
            ResourcesNotAvailableError: If the requested resources exceed the
                available resources.
        """

        if aggregator.private_attributes and aggregator.private_attributes_callable:
            self.logger.warning(
                "Warning: Aggregator private attributes "
                + "will be initialized via callable and "
                + "attributes via aggregator.private_attributes "
                + "will be ignored"
            )

        if self.backend == "single_process":
            return aggregator

        total_available_cpus = os.cpu_count()
        total_available_gpus = get_number_of_gpus()

        agg_cpus = aggregator.num_cpus
        agg_gpus = aggregator.num_gpus

        if agg_gpus > 0:
            check_resource_allocation(
                total_available_gpus,
                {aggregator.get_name(): agg_gpus},
            )

        if total_available_gpus < agg_gpus:
            raise ResourcesNotAvailableError(
                f"cannot assign more than available GPUs \
                    ({agg_gpus} < {total_available_gpus})."
            )
        if total_available_cpus < agg_cpus:
            raise ResourcesNotAvailableError(
                f"cannot assign more than available CPUs \
                    ({agg_cpus} < {total_available_cpus})."
            )

        interface_module = importlib.import_module("openfl.experimental.workflow.interface")
        aggregator_class = interface_module.Aggregator

        aggregator_actor = ray.remote(aggregator_class).options(
            num_cpus=agg_cpus, num_gpus=agg_gpus
        )

        if aggregator.private_attributes_callable is not None:
            aggregator_actor_ref = aggregator_actor.remote(
                name=aggregator.get_name(),
                private_attributes_callable=aggregator.private_attributes_callable,
                **aggregator.kwargs,
            )
        elif aggregator.private_attributes is not None:
            aggregator_actor_ref = aggregator_actor.remote(
                name=aggregator.get_name(),
                **aggregator.kwargs,
            )
            aggregator_actor_ref.initialize_private_attributes.remote(aggregator.private_attributes)

        return aggregator_actor_ref

    def __get_collaborator_object(self, collaborators: List) -> Any:
        """Get collaborator object based on localruntime backend.

        If the backend is 'single_process', it returns the list of
        collaborators directly.
        If the backend is 'ray', it assigns collaborators to Ray actors using
        the ray_group_assign function.

        Args:
            collaborators (List[Type[Collaborator]]): The list of collaborator
                classes to instantiate.

        Returns:
            Any: The list of collaborator objects or a list of references to
                the Ray actors representing the collaborators.

        Raises:
            ResourcesNotAvailableError: If the requested resources exceed the
                available resources.
        """
        for collab in collaborators:
            if collab.private_attributes and collab.private_attributes_callable:
                self.logger.warning(
                    f"Warning: Collaborator {collab.name} private attributes "
                    + "will be initialized via callable and "
                    + "attributes via collaborator.private_attributes "
                    + "will be ignored"
                )

        if self.backend == "single_process":
            return collaborators

        total_available_cpus = os.cpu_count()
        total_required_cpus = sum([collaborator.num_cpus for collaborator in collaborators])
        if total_available_cpus < total_required_cpus:
            raise ResourcesNotAvailableError(
                f"cannot assign more than available CPUs \
                    ({total_required_cpus} < {total_available_cpus})."
            )

        if self.backend == "ray":
            collaborator_ray_refs = ray_group_assign(collaborators, num_actors=self.num_actors)
            return collaborator_ray_refs

    @property
    def aggregator(self) -> str:
        """Gets the name of the aggregator.

        Returns:
            str: The name of the aggregator.
        """
        return self._aggregator.name

    @aggregator.setter
    def aggregator(self, aggregator: Type[Aggregator]):
        """Set LocalRuntime _aggregator.

        Args:
            aggregator (Type[Aggregator]): The aggregator to be set.
        """
        self._aggregator = aggregator

    @property
    def collaborators(self) -> List[str]:
        """Return names of collaborators. Don't give direct access to private
        attributes.

        Returns:
            List[str]: The names of the collaborators.
        """
        return list(self.__collaborators.keys())

    @collaborators.setter
    def collaborators(self, collaborators: List[Type[Collaborator]]):
        """Set LocalRuntime collaborators.

        Args:
            collaborators (List[Type[Collaborator]]): The collaborators to be
                set.
        """
        if self.backend == "single_process":

            def get_collab_name(collab):
                return collab.get_name()

        else:

            def get_collab_name(collab):
                return ray.get(collab.get_name.remote())

        self.__collaborators = {
            get_collab_name(collaborator): collaborator for collaborator in collaborators
        }

    def get_collaborator_kwargs(self, collaborator_name: str):
        """Returns kwargs of collaborator.

        Args:
            collaborator_name: Collaborator name for which kwargs is to be
                returned

        Returns:
            kwargs: Collaborator private_attributes_callable function name, and
             arguments required to call it.
        """
        collab = self.__collaborators[collaborator_name]
        kwargs = {}
        if hasattr(collab, "private_attributes_callable"):
            if collab.private_attributes_callable is not None:
                kwargs.update(collab.kwargs)
                kwargs["private_attributes_callable"] = collab.private_attributes_callable.__name__

        return kwargs

    def initialize_aggregator(self):
        """Initialize aggregator private attributes."""
        if self.backend == "single_process":
            self._aggregator.initialize_private_attributes()
        else:
            ray.get(self._aggregator.initialize_private_attributes.remote())

    def initialize_collaborators(self):
        """Initialize collaborator private attributes."""
        if self.backend == "single_process":

            def init_private_attrs(collab):
                return collab.initialize_private_attributes()

        else:

            def init_private_attrs(collab):
                return ray.get(collab.initialize_private_attributes.remote())

        for collaborator in self.__collaborators.values():
            init_private_attrs(collaborator)

    def restore_instance_snapshot(self, ctx: Type[FLSpec], instance_snapshot: List[Type[FLSpec]]):
        """Restores attributes from backup (in instance snapshot) to context
        (ctx).

        Args:
            ctx (Type[FLSpec]): The context to restore the snapshot to.
            instance_snapshot (List[Type[FLSpec]]): The snapshot of the
                instance to be restored.
        """
        for backup in instance_snapshot:
            artifacts_iter, _ = generate_artifacts(ctx=backup)
            for name, attr in artifacts_iter():
                if not hasattr(ctx, name):
                    setattr(ctx, name, attr)

    def execute_agg_steps(self, ctx: Any, f_name: str, clones: Optional[Any] = None):
        """
        Execute aggregator steps until at transition point.

        Args:
            ctx (Any): The context in which the function is executed.
            f_name (str): The name of the function to be executed.
            clones (Optional[Any], optional): Clones if any. Defaults to None.
        """
        if clones is not None:
            f = getattr(ctx, f_name)
            f(clones)
        else:
            not_at_transition_point = True
            while not_at_transition_point:
                f = getattr(ctx, f_name)
                f()

                f, parent_func = ctx.execute_task_args[:2]
                if aggregator_to_collaborator(f, parent_func) or f.__name__ == "end":
                    not_at_transition_point = False

                f_name = f.__name__

    def execute_collab_steps(self, ctx: Any, f_name: str):
        """Execute collaborator steps until at transition point.

        Args:
            ctx (Any): The context in which the function is executed.
            f_name (str): The name of the function to be executed.
        """
        not_at_transition_point = True
        while not_at_transition_point:
            f = getattr(ctx, f_name)
            f()

            f, parent_func = ctx.execute_task_args[:2]
            if ctx._is_at_transition_point(f, parent_func):
                not_at_transition_point = False

            f_name = f.__name__

    def execute_task(self, flspec_obj: Type[FLSpec], f: Callable, **kwargs):
        """Defines which function to be executed based on name and kwargs.

        Updates the arguments and executes until end is not reached.

        Args:
            flspec_obj: Reference to the FLSpec (flow) object. Contains
                information about task sequence, flow attributes.
            f: The next task to be executed within the flow.

        Returns:
            artifacts_iter: Iterator with updated sequence of values
        """
        parent_func = None
        instance_snapshot = None
        self.join_step = False

        while f.__name__ != "end":
            if "foreach" in kwargs:
                flspec_obj = self.execute_collab_task(
                    flspec_obj, f, parent_func, instance_snapshot, **kwargs
                )
            else:
                flspec_obj = self.execute_agg_task(flspec_obj, f)
            f, parent_func, instance_snapshot, kwargs = flspec_obj.execute_task_args
        else:
            flspec_obj = self.execute_agg_task(flspec_obj, f)
            f = flspec_obj.execute_task_args[0]

            checkpoint(flspec_obj, f)
            artifacts_iter, _ = generate_artifacts(ctx=flspec_obj)
            return artifacts_iter()

    def execute_agg_task(self, flspec_obj, f):
        """Performs execution of aggregator task.

        Args:
            flspec_obj: Reference to the FLSpec (flow) object.
            f: The task to be executed within the flow.

        Returns:
            flspec_obj: updated FLSpec (flow) object.
        """

        aggregator = self._aggregator
        clones = None

        if self.join_step:
            clones = [FLSpec._clones[col] for col in self.selected_collaborators]
            self.join_step = False

        if self.backend == "ray":
            ray_executor = RayExecutor()
            ray_executor.ray_call_put(
                aggregator,
                flspec_obj,
                f.__name__,
                self.execute_agg_steps,
                clones,
            )
            flspec_obj = ray_executor.ray_call_get()[0]
            del ray_executor
        else:
            aggregator.execute_func(flspec_obj, f.__name__, self.execute_agg_steps, clones)

        gc.collect()
        return flspec_obj

    def execute_collab_task(self, flspec_obj, f, parent_func, instance_snapshot, **kwargs):
        """
        Performs
            1. Filter include/exclude
            2. Set runtime, collab private attributes , metaflow_interface
            3. Execution of all collaborator for each task
            4. Remove collaborator private attributes
            5. Execute the next function after transition

        Args:
            flspec_obj: Reference to the FLSpec (flow) object.
            f: The task to be executed within the flow.
            parent_func: The prior task executed in the flow.
            instance_snapshot: A prior FLSpec state that needs to be restored.

        Returns:
            flspec_obj: updated FLSpec (flow) object
        """

        flspec_obj._foreach_methods.append(f.__name__)
        selected_collaborators = getattr(flspec_obj, kwargs["foreach"])
        self.selected_collaborators = selected_collaborators

        # filter exclude/include attributes for clone
        self.filter_exclude_include(flspec_obj, f, selected_collaborators, **kwargs)

        if self.backend == "ray":
            ray_executor = RayExecutor()
        # set runtime,collab private attributes and metaflowinterface
        for col in selected_collaborators:
            clone = FLSpec._clones[col]
            # Set new LocalRuntime for clone as it is required
            # new runtime object will not contain private attributes of
            # aggregator or other collaborators
            clone.runtime = LocalRuntime(backend="single_process")

            # write the clone to the object store
            # ensure clone is getting latest _metaflow_interface
            clone._metaflow_interface = flspec_obj._metaflow_interface

        for collab_name in selected_collaborators:
            clone = FLSpec._clones[collab_name]
            collaborator = self.__collaborators[collab_name]

            if self.backend == "ray":
                ray_executor.ray_call_put(
                    collaborator, clone, f.__name__, self.execute_collab_steps
                )
            else:
                collaborator.execute_func(clone, f.__name__, self.execute_collab_steps)

        if self.backend == "ray":
            clones = ray_executor.ray_call_get()
            FLSpec._clones.update(zip(selected_collaborators, clones))
            clone = clones[0]
            del clones

        flspec_obj.execute_task_args = clone.execute_task_args

        # Restore the flspec_obj state if back-up is taken
        self.restore_instance_snapshot(flspec_obj, instance_snapshot)
        del instance_snapshot

        gc.collect()
        # Setting the join_step to indicate to aggregator to collect clones
        self.join_step = True
        return flspec_obj

    def filter_exclude_include(self, flspec_obj, f, selected_collaborators, **kwargs):
        """
        This function filters exclude/include attributes
        Args:
            flspec_obj: Reference to the FLSpec (flow) object.
            f: The task to be executed within the flow.
            selected_collaborators: all collaborators.
        """

        for col in selected_collaborators:
            clone = FLSpec._clones[col]
            clone.input = col
            if ("exclude" in kwargs and hasattr(clone, kwargs["exclude"][0])) or (
                "include" in kwargs and hasattr(clone, kwargs["include"][0])
            ):
                filter_attributes(clone, f, **kwargs)
            artifacts_iter, _ = generate_artifacts(ctx=flspec_obj)
            for name, attr in artifacts_iter():
                setattr(clone, name, deepcopy(attr))
            clone._foreach_methods = flspec_obj._foreach_methods

    def __repr__(self):
        """Returns the string representation of the LocalRuntime object.

        Returns:
            str: The string representation of the LocalRuntime object.
        """
        return "LocalRuntime"



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Experimental Collaborator module."""

import time
from logging import getLogger
from typing import Any, Callable, Dict, Tuple

import dill


class Collaborator:
    r"""The Collaborator object class.

    Args:
        collaborator_name (str): The common name for the collaborator.
        aggregator_uuid (str): The unique id for the client.
        federation_uuid (str): The unique id for the federation.

        client (AggregatorGRPCClient): GRPC Client to connect to
        Aggregator Server.

        private_attrs_callable (Callable): Function for Collaborator
        private attributes.
        private_attrs_kwargs (Dict): Arguments to call private_attrs_callable.

    Note:
        \* - Plan setting.
    """

    def __init__(
        self,
        collaborator_name: str,
        aggregator_uuid: str,
        federation_uuid: str,
        client: Any,
        private_attributes_callable: Any = None,
        private_attributes_kwargs: Dict = None,
        private_attributes: Dict = None,
        **kwargs,
    ) -> None:
        self.name = collaborator_name
        self.aggregator_uuid = aggregator_uuid
        self.federation_uuid = federation_uuid

        self.client = client

        self.logger = getLogger(__name__)

        self.__private_attrs_callable = private_attributes_callable

        self.__private_attrs = private_attributes
        if self.__private_attrs_callable is not None:
            self.logger.info("Initializing collaborator.")
            self.__initialize_private_attributes(private_attributes_kwargs)

    def __initialize_private_attributes(self, kwargs: Dict) -> None:
        """Call private_attrs_callable function set attributes to
        self.__private_attrs.

        Args:
            kwargs (Dict): Private attributes callable function arguments

        Returns:
            None
        """
        self.__private_attrs = self.__private_attrs_callable(**kwargs)

    def __set_private_attrs_to_clone(self, clone: Any) -> None:
        """Set private_attrs of Collaborator as attributes of FLSpec clone.

        Args:
            clone (FLSpec): Clone to which private attributes are to be
            set

        Returns:
            None
        """
        if len(self.__private_attrs) > 0:
            for name, attr in self.__private_attrs.items():
                setattr(clone, name, attr)

    def __delete_private_attrs_from_clone(self, clone: Any, replace_str: str = None) -> None:
        """
        Remove collaborator private attributes from FLSpec clone
        before transition from Collaborator step to aggregator steps.
        Instead of removing private attributes this method can also replace
        private attributes with a string (required in checkpointing)

        Args:
            clone (FLSpec): Clone from which private attributes are to be
            removed

        Returns:
            None
        """
        # Update aggregator private attributes by taking latest
        # parameters from clone, then delete attributes from clone.
        if len(self.__private_attrs) > 0:
            for attr_name in self.__private_attrs:
                if hasattr(clone, attr_name):
                    self.__private_attrs.update({attr_name: getattr(clone, attr_name)})
                    if replace_str:
                        setattr(clone, attr_name, replace_str)
                    else:
                        delattr(clone, attr_name)

    def call_checkpoint(self, ctx: Any, f: Callable, stream_buffer: Any) -> None:
        """
        Call checkpoint gRPC.

        Args:
            ctx (FLSpec): FLSPec object.
            f (Callable): Flow step which is be checkpointed.
            stream_buffer (Any): Captured object for output and error.

        Returns:
            None
        """
        self.client.call_checkpoint(
            self.name,
            dill.dumps(ctx),
            dill.dumps(f),
            dill.dumps(stream_buffer),
        )

    def run(self) -> None:
        """Run the collaborator.

        Args:
            None

        Returns:
            None
        """
        while True:
            next_step, clone, sleep_time, time_to_quit = self.get_tasks()
            if time_to_quit:
                break
            elif sleep_time > 0:
                time.sleep(sleep_time)
            else:
                self.logger.info(f"Received the following tasks: {next_step}.")
                f_name, ctx = self.do_task(next_step, clone)
                self.send_task_results(f_name, ctx)

        self.logger.info("End of Federation reached. Exiting...")

    def send_task_results(self, next_step: str, clone: Any) -> None:
        """After collaborator is executed, send next aggregator step to
        Aggregator for continue execution.

        Args:
            next_step (str): Send next function to aggregator
            clone (FLSpec): Updated clone object (Private attributes atr not
                included)

        Returns:
            None
        """
        self.logger.info(
            f"Round {self.round_number}, collaborator {self.name} is sending results..."
        )
        self.client.send_task_results(self.name, self.round_number, next_step, dill.dumps(clone))

    def get_tasks(self) -> Tuple:
        """Get tasks from the aggregator.

        Args:
            None

        Returns:
            next_step (str): Next collaborator function to start execution from
            ctx (FLSpec): Function context
            sleep_time (int): Sleep for given seconds if not ready yet
            time_to_quit (bool): True if end of reached
        """
        self.logger.info("Waiting for tasks...")
        temp = self.client.get_tasks(self.name)
        self.round_number, next_step, clone_bytes, sleep_time, time_to_quit = temp
        if time_to_quit:
            return next_step, "", sleep_time, time_to_quit
        return next_step, dill.loads(clone_bytes), sleep_time, time_to_quit

    def do_task(self, f_name: str, ctx: Any) -> Tuple:
        """Run collaborator steps until transition.

        Args:
            f_name (str): Function name which is to be executed.
            ctx (FLSpec): Function context.

        Returns:
            Tuple(str, FLSpec): Next aggregator function, and updated context.
        """
        # Set private attributes to context
        self.__set_private_attrs_to_clone(ctx)

        # Loop control variable
        not_at_transition_point = True
        while not_at_transition_point:
            f = getattr(ctx, f_name)
            f()
            # Checkpoint the function
            self.__delete_private_attrs_from_clone(ctx, "Private attributes: Not Available.")
            self.call_checkpoint(ctx, f, f._stream_buffer)
            self.__set_private_attrs_to_clone(ctx)

            _, f, parent_func = ctx.execute_task_args[:3]
            # Display transition logs if transition
            ctx._display_transition_logs(f, parent_func)

            # If transition break the loop
            if ctx._is_at_transition_point(f, parent_func):
                not_at_transition_point = False

            # Update the function name
            f_name = f.__name__

        # Remove private attributes from context
        self.__delete_private_attrs_from_clone(ctx)

        return f_name, ctx



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""openfl.experimental.workflow.interface.flspec module."""

from __future__ import annotations

import inspect
from copy import deepcopy
from typing import TYPE_CHECKING, Callable, List, Type, Union

if TYPE_CHECKING:
    from openfl.experimental.workflow.runtime import FederatedRuntime, LocalRuntime, Runtime

from openfl.experimental.workflow.utilities import (
    MetaflowInterface,
    SerializationError,
    aggregator_to_collaborator,
    checkpoint,
    collaborator_to_aggregator,
    filter_attributes,
    generate_artifacts,
    should_transfer,
)


class FLSpec:
    """FLSpec Class

    A class representing a Federated Learning Specification. It manages clones,
    maintains the initial state, and supports checkpointing.

    Attributes:
        _clones (list): A list of clones created for the FLSpec instance.
        _initial_state (FLSpec or None): The saved initial state of the FLSpec instance.
        _foreach_methods (list): A list of methods to be applied iteratively.
        _checkpoint (bool): A flag indicating whether checkpointing is enabled.
        _runtime (RuntimeType): The runtime of the flow.
    """

    _clones = []
    _initial_state = None

    def __init__(self, checkpoint: bool = False) -> None:
        """Initializes the FLSpec object.

        Args:
            checkpoint (bool, optional): Determines whether to checkpoint or
                not. Defaults to False.
        """
        self._foreach_methods = []
        self._checkpoint = checkpoint

    @classmethod
    def _create_clones(cls, instance: Type[FLSpec], names: List[str]) -> None:
        """Creates clones for instance for each collaborator in names.

        Args:
            instance (Type[FLSpec]): The instance to be cloned.
            names (List[str]): The list of names for the clones.
        """
        cls._clones = {name: deepcopy(instance) for name in names}

    @classmethod
    def _reset_clones(cls) -> None:
        """Resets the clones of the class."""

        cls._clones = []

    @classmethod
    def save_initial_state(cls, instance: Type[FLSpec]) -> None:
        """Saves the initial state of an instance before executing the flow.

        Args:
            instance (Type[FLSpec]): The instance whose initial state is to be
                saved.
        """
        cls._initial_state = deepcopy(instance)

    @property
    def checkpoint(self) -> bool:
        """Getter for the checkpoint attribute.

        Returns:
            bool: The current value of the checkpoint.
        """
        return self._checkpoint

    @checkpoint.setter
    def checkpoint(self, value: bool) -> None:
        """Setter for the checkpoint attribute.

        Args:
            value (bool): The new value for the checkpoint.

        Raises:
            ValueError: If the provided value is not a boolean.
        """
        if not isinstance(value, bool):
            raise ValueError("checkpoint must be a boolean value.")
        self._checkpoint = value

    @property
    def runtime(self) -> Type[Union[LocalRuntime, FederatedRuntime]]:
        """Returns flow runtime.

        Returns:
            Type[Runtime]: The runtime of the flow.
        """
        return self._runtime

    @runtime.setter
    def runtime(self, runtime: Type[Runtime]) -> None:
        """Sets flow runtime.

        Args:
            runtime (Type[Runtime]): The runtime to be set.

        Raises:
            TypeError: If the provided runtime is not a valid OpenFL Runtime.
        """
        if str(runtime) not in ["LocalRuntime", "FederatedRuntime"]:
            raise TypeError(f"{runtime} is not a valid OpenFL Runtime")
        self._runtime = runtime

    def run(self) -> None:
        """Starts the execution of the flow."""
        # Submit flow to Runtime
        if str(self._runtime) == "LocalRuntime":
            self._run_local()
        elif str(self._runtime) == "FederatedRuntime":
            self._run_federated()
        else:
            raise Exception("Runtime not implemented")

    def _run_local(self) -> None:
        """Executes the flow using LocalRuntime."""
        self._setup_initial_state()
        try:
            # Execute all Participant (Aggregator & Collaborator) tasks and
            # retrieve the final attributes
            # start step is the first task & invoked on aggregator through
            # runtime.execute_task
            final_attributes = self.runtime.execute_task(
                self,
                self.start,
            )
        except Exception as e:
            if "cannot pickle" in str(e) or "Failed to unpickle" in str(e):
                msg = (
                    "\nA serialization error was encountered that could not"
                    "\nbe handled by the ray backend."
                    "\nTry rerunning the flow without ray as follows:\n"
                    "\nLocalRuntime(...,backend='single_process')\n"
                    "\n or for more information about the original error,"
                    "\nPlease see the official Ray documentation"
                    "\nhttps://docs.ray.io/en/releases-2.2.0/ray-core/\
                    objects/serialization.html"
                )
                raise SerializationError(str(e) + msg)
            else:
                raise e
        for name, attr in final_attributes:
            setattr(self, name, attr)

    def _setup_initial_state(self) -> None:
        """
        Sets up the flow's initial state, initializing private attributes for
        collaborators and aggregators.
        """
        self._metaflow_interface = MetaflowInterface(self.__class__, self.runtime.backend)
        self._run_id = self._metaflow_interface.create_run()
        # Initialize aggregator private attributes
        self.runtime.initialize_aggregator()
        self._foreach_methods = []
        FLSpec._reset_clones()
        FLSpec._create_clones(self, self.runtime.collaborators)
        # Initialize collaborator private attributes
        self.runtime.initialize_collaborators()
        if self._checkpoint:
            print(f"Created flow {self.__class__.__name__}")

    def _run_federated(self) -> None:
        """Executes the flow using FederatedRuntime."""
        try:
            # Prepare workspace and submit it for the FederatedRuntime
            archive_path, exp_name = self.runtime.prepare_workspace_archive()
            self.runtime.submit_experiment(archive_path, exp_name)
            # Stream the experiment's stdout if the checkpoint is enabled
            if self._checkpoint:
                self.runtime.stream_experiment_stdout(exp_name)
            # Retrieve the flspec object to update the experiment state
            flspec_obj = self._get_flow_state()
            # Update state of self
            self._update_from_flspec_obj(flspec_obj)
        except Exception as e:
            raise Exception(
                f"FederatedRuntime: Experiment {exp_name} failed to run due to error: {e}"
            )

    def _update_from_flspec_obj(self, flspec_obj: FLSpec) -> None:
        """Update self with attributes from the updated flspec instance.

        Args:
            flspec_obj (FLSpec): Updated Flspec instance
        """
        artifacts_iter, _ = generate_artifacts(ctx=flspec_obj)
        for name, attr in artifacts_iter():
            setattr(self, name, deepcopy(attr))

        self._foreach_methods = flspec_obj._foreach_methods

    def _get_flow_state(self) -> Union[FLSpec, None]:
        """
        Gets the updated flow state.

        Returns:
            flspec_obj (Union[FLSpec, None]): An updated FLSpec instance if the experiment
                runs successfully. None if the experiment could not run.
        """
        status, flspec_obj = self.runtime.get_flow_state()
        if status:
            print("Experiment ran successfully")
            return flspec_obj
        else:
            print("Experiment could not run")
            return None

    def _capture_instance_snapshot(self, kwargs) -> List:
        """Takes backup of self before exclude or include filtering.

        Args:
            kwargs: Key word arguments originally passed to the next function.
                    If include or exclude are in the kwargs, the state of the
                    aggregator needs to be retained.

        Returns:
            return_objs (list): A list of return objects.
        """
        return_objs = []
        if "exclude" in kwargs or "include" in kwargs:
            backup = deepcopy(self)
            return_objs.append(backup)
        return return_objs

    def _is_at_transition_point(self, f: Callable, parent_func: Callable) -> bool:
        """
        Determines if the collaborator has finished its current sequence.

        Args:
            f (Callable): The next function to be executed.
            parent_func (Callable): The previous function executed.

        Returns:
            bool: True if the collaborator has finished its current sequence,
                False otherwise.
        """
        if parent_func.__name__ in self._foreach_methods:
            self._foreach_methods.append(f.__name__)
            if should_transfer(f, parent_func):
                print(f"Should transfer from {parent_func.__name__} to {f.__name__}")
                self.execute_next = f.__name__
                return True
        return False

    def _display_transition_logs(self, f: Callable, parent_func: Callable) -> None:
        """
        Prints aggregator to collaborators or collaborators to aggregator
        state transition logs.

        Args:
            f (Callable): The next function to be executed.
            parent_func (Callable): The previous function executed.
        """
        if aggregator_to_collaborator(f, parent_func):
            print("Sending state from aggregator to collaborators")

        elif collaborator_to_aggregator(f, parent_func):
            print("Sending state from collaborator to aggregator")

    def filter_exclude_include(self, f, **kwargs) -> None:
        """Filters exclude/include attributes for a given task within the flow.

        Args:
            f (Callable): The task to be executed within the flow.
            **kwargs (dict): Additional keyword arguments. These should
                include:
                - "foreach" (str): The attribute name that contains the list
                of selected collaborators.
                - "exclude" (list, optional): List of attribute names to
                exclude. If an attribute name is present in this list and the
                clone has this attribute, it will be filtered out.
                - "include" (list, optional): List of attribute names to
                include. If an attribute name is present in this list and the
                clone has this attribute, it will be included.
        """
        selected_collaborators = getattr(self, kwargs["foreach"])

        for col in selected_collaborators:
            clone = FLSpec._clones[col]
            clone.input = col
            if ("exclude" in kwargs and hasattr(clone, kwargs["exclude"][0])) or (
                "include" in kwargs and hasattr(clone, kwargs["include"][0])
            ):
                filter_attributes(clone, f, **kwargs)
            artifacts_iter, _ = generate_artifacts(ctx=self)
            for name, attr in artifacts_iter():
                setattr(clone, name, deepcopy(attr))
            clone._foreach_methods = self._foreach_methods

    def restore_instance_snapshot(self, ctx: FLSpec, instance_snapshot: List[FLSpec]) -> None:
        """Restores attributes from backup (in instance snapshot) to ctx.

        Args:
            ctx (FLSpec): The context to restore the attributes to.
            instance_snapshot (List[FLSpec]): The list of FLSpec instances
                that serve as the backup.
        """
        for backup in instance_snapshot:
            artifacts_iter, _ = generate_artifacts(ctx=backup)
            for name, attr in artifacts_iter():
                if not hasattr(ctx, name):
                    setattr(ctx, name, attr)

    def next(self, f, **kwargs) -> None:
        """Specifies the next task in the flow to execute.

        Args:
            f (Callable): The next task that will be executed in the flow.
            **kwargs: Additional keyword arguments.
        """
        # Get the name and reference to the calling function
        parent = inspect.stack()[1][3]
        parent_func = getattr(self, parent)

        if str(self._runtime) == "LocalRuntime":
            # Checkpoint current attributes (if checkpoint==True)
            checkpoint(self, parent_func)

        # Take back-up of current state of self
        agg_to_collab_ss = None
        if aggregator_to_collaborator(f, parent_func):
            agg_to_collab_ss = self._capture_instance_snapshot(kwargs=kwargs)

        # Remove included / excluded attributes from next task
        filter_attributes(self, f, **kwargs)

        if str(self._runtime) == "FederatedRuntime":
            if f.collaborator_step and not f.aggregator_step:
                self._foreach_methods.append(f.__name__)

            self.execute_task_args = (
                self,
                f,
                parent_func,
                FLSpec._clones,
                agg_to_collab_ss,
                kwargs,
            )

        elif str(self._runtime) == "LocalRuntime":
            # update parameters required to execute execute_task function
            self.execute_task_args = [f, parent_func, agg_to_collab_ss, kwargs]



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Detect criteria for transitions in placement."""


def should_transfer(func, parent_func):
    """Determines if a transfer should occur from collaborator to aggregator.

    Args:
        func (function): The current function.
        parent_func (function): The parent function.

    Returns:
        bool: True if a transfer should occur, False otherwise.
    """
    if collaborator_to_aggregator(func, parent_func):
        return True
    else:
        return False


def aggregator_to_collaborator(func, parent_func):
    """Checks if a transition from aggregator to collaborator is possible.

    Args:
        func (function): The current function.
        parent_func (function): The parent function.

    Returns:
        bool: True if the transition is possible, False otherwise.
    """
    if parent_func.aggregator_step and func.collaborator_step:
        return True
    else:
        return False


def collaborator_to_aggregator(func, parent_func):
    """Checks if a transition from collaborator to aggregator is possible.

    Args:
        func (function): The current function.
        parent_func (function): The parent function.

    Returns:
        bool: True if the transition is possible, False otherwise.
    """
    if parent_func.collaborator_step and func.aggregator_step:
        return True
    else:
        return False



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Mock objects to eliminate extraneous dependencies"""


class MockDataLoader:
    """Placeholder dataloader for when data is not available"""

    def __init__(self, feature_shape):
        self.feature_shape = feature_shape

    def get_feature_shape(self):
        return self.feature_shape

    def get_train_data_size(self):
        return 0

    def get_valid_data_size(self):
        return 0
